# -*- coding: utf-8 -*-

# romiq.kh@gmail.com, 2014

import array, struct, io

from . import EngineError

class MSKLoader:
    def __init__(self):
        self.bound = [0, 0, 0, 0]
        self.rects = []

    def load_info(self, f):
        return self.load_data(f)

    def load_data(self, f):
        rest = f.read()
        f.seek(0)
        delta = len(rest) - 16
        rects = []
        while delta > 0:
            temp = f.read(4)
            delta -= 4
            rects_len = struct.unpack_from("<I", temp)[0]
            rec = []
            for r_ref in range(rects_len):
                temp = f.read(8)
                delta -= 8
                l,t,r,b = struct.unpack_from("<4h", temp)
                rec.append([l, t, r, b])
            rects.append(rec)
            delta -= 4

        if delta != 0:
           raise EngineError("Bad MSK file")

        temp = f.read(len(rects) * 4)
        frms = struct.unpack_from("<{}I".format(len(rects)), temp)

        if len(rects) != len(frms):
            raise EngineError("Bad MSK file structure")

        temp = f.read(16)
        self.bound = struct.unpack_from("<4i", temp)

        self.rects = list(zip(reversed(frms), rects))



# -*- coding: utf-8 -*-

# romiq.kh@gmail.com, 2014

import os
import struct
import io

from . import EngineError

# manage files data
class FileManager:
    def __init__(self, root):
        self.root = os.path.abspath(root)

        self.strfd = []
        self.strtable = {}
        self.strtableord = []

    def find_path(self, path):
        # search case insensive from root
        dpath = []
        npath = self.root
        path = path.replace("\\", "/")
        for item in path.split("/"):
            if not item: continue
            ok = False
            for ritem in os.listdir(npath):
                if item.lower() != ritem.lower(): continue
                npath = os.path.join(npath, ritem)
                ok = True
                break
            if not ok: return None
        return npath

    def load_store(self, name, tag = 0):
        path = self.find_path(name)
        if path is None:
            print("DEBUG: Store \"{}\" not found".format(name))
            return
        # scan table
        f = open(path, "rb")
        # check magic string "StOR"
        magic = f.read(4)
        if magic != b"StOR":
            raise EngineError("Bad magic in store \"{}\"".format(name))
            return
        # read index table ref
        temp = f.read(4)
        index_ref = struct.unpack_from("<I", temp)[0]
        f.seek(index_ref)
        # index table length
        temp = f.read(4)
        index_len = struct.unpack_from("<I", temp)[0]
        index_table = []
        for iref in range(index_len):
            temp = f.read(12)
            data = struct.unpack_from("<III", temp)
            index_table.append((data[1], data[2]))
        strlst = []
        data = f.read().decode("latin-1")
        for idx, fname in enumerate(data.split("\x00")):
            fname = fname.lower().replace("\\", "/")
            if idx < index_len and fname not in self.strtable:
                self.strtable[fname] = (len(self.strfd),) + index_table[idx]
                strlst.append((fname, len(self.strtableord)) + index_table[idx])
                self.strtableord.append(fname)
            else:
                if len(fname) > 0:
                    print("DEBUG:Extra file record \"{}\" in \"{}\"".\
                        format(fname, name))
        # add file descriptor
        self.strfd.append((f, name, tag, strlst))
        print("DEBUG: Loaded store \"{}\"".format(name))

    def read_file(self, fname):
        sf = fname.lower().replace("\\", "/")
        if sf in self.strtable:
            fnum, st, ln = self.strtable[sf]
            print("Load file \"{}\" from store \"{}\"".\
                format(fname, self.strfd[fnum][1]))
            self.strfd[fnum][0].seek(st)
            return self.strfd[fnum][0].read(ln)
        else:
            print("Load file \"{}\" from filesystem".format(fname))
            pf = self.find_path(fname)
            if not pf:
                print("DEBUG: Can't open file \"{}\"".format(fname))
            # file in filesystem
            f = open(pf, "rb")
            try:
                data = f.read()
            finally:
                f.close()
            return data

    def read_file_stream(self, fname):
        data = self.read_file(fname)
        mems = io.BytesIO()
        mems.write(data)
        mems.seek(0)
        return mems

    def exists(self, fname):
        sf = fname.lower().replace("\\", "/")
        if sf in self.strtable:
            return True
        else:
            return self.find_path(fname) is not None

    def unload_stores(self, flt = None):
        strfd = []
        strtable = {}
        strtableord = []
        for idx, (fd, name, tag, strlst) in enumerate(self.strfd):
            if flt is not None:
                if tag != flt:
                    for k, v in self.strtable.items():
                        if v[0] == idx:
                            strtable[k] = (len(strfd), v[1], v[2])
                            strtableord.append(k)
                    strfd.append((fd, name, tag, strlst))
                    continue
            print("DEBUG: Unload store \"{}\"".format(name))
            try:
                if fd: fd.close()
            except Exception as e:
                print("DEBUG: Can't unload \"{}\":".format(name) + str(e))
        self.strfd = strfd
        self.strtable = strtable
        self.strtableord = strtableord



# -*- coding: utf-8 -*-

# romiq.kh@gmail.com, 2014

import array, struct, io

from . import EngineError

class LEGLoader:
    def __init__(self):
        self.coords = []

    def load_info(self, f):
        return self.load_data(f)

    def load_data(self, f):
        hdr = f.read(4)
        if hdr != b"xyof":
            raise EngineError("Bad LEG/OFF magic \"{}\"".format(hdr))

        rest = f.read()
        if len(rest) % 8:
            raise EngineError("Bad LEG/OFF size {}".format(len(rest)))

        sf = struct.unpack("<{}l".format(len(rest) // 4), rest)
        self.coords = [[sf[i * 2], sf[i * 2 + 1]] for i in range(len(rest) // 8)]



# -*- coding: utf-8 -*-

# romiq.kh@gmail.com, 2014

import array, struct, io
import binascii

from . import EngineError, BMPLoader


class SaveLoader:
    def __init__(self, enc = None):
        self.part = None
        self.chap = None
        self.stamp = None
        self.enc = enc
        self.objects = []

    def load_data(self, f, part, objnum):
        data = f.read(8)
        self.part, self.chap = struct.unpack("<2I", data)
        if self.part != part: return

        # read date stamp, asciiz
        stamp = f.read(30)
        stamp = stamp.split(b"\x00")[0]
        self.stamp = stamp.decode(self.enc)

        # read screenshot
        sl = 108 * 81 * 2
        rgb = f.read(sl)
        if len(rgb) != sl:
            raise EngineError("Bad SAVE length (no screenshot)")
        self.shot = BMPLoader()
        self.shot.load_raw(108, 81, rgb)

        hz2 = f.read(216)
        if hz2 != b"\x00" * 216:
            raise EngineError("Bad SAVE error in HZ2 field")

        data = f.read(4)
        hz3 = struct.unpack("<I", data)[0]

        if hz3 != objnum + 3:
            raise EngineError("Bad SAVE objects number")

        def readstr():
            data = f.read(4)
            strlen = struct.unpack("<I", data)[0]
            s = f.read(strlen)
            #print("STR", strlen, data, s)
            return s.decode(self.enc)

        for i in range(objnum):
            s1 = readstr()
            s2 = readstr()
            data = f.read(33)
            obj = {"name": s1, "alias": s2, "data": data}
            recs = struct.unpack("<iB7i", data)
            obj["recs"] = recs
            obj["res"] = recs[2]
            self.objects.append(obj)
            #print(i, s1, s2)

        # invntr
        data = f.read(4)
        invlen = struct.unpack("<I", data)[0]
        data = f.read(invlen * 2)
        self.invntr = struct.unpack("<{}H".format(invlen), data)

        # scene
        self.scene = readstr()

        # char positions
        data = f.read(16)
        charpos = struct.unpack("<4I", data)

        # arr dialog opcodes
        data = f.read(4)
        dlgoplen = struct.unpack("<I", data)[0]
        self.dlgops = []
        for _ in range(dlgoplen):
            data = f.read(4)
            ref, arg, code = struct.unpack_from("<HBB", data)
            self.dlgops.append([code, arg, ref])

        data = f.read(20)
        self.cursor_res, self.cursor, self.cursor_obj, c1res, c2res = \
            struct.unpack("<5I", data)

        # charters: x, y, res
        self.char1 = (charpos[0], charpos[1], c1res)
        self.char2 = (charpos[2], charpos[3], c2res)

        hz7 = f.read(32)
        if hz7 != b"\xff" * 32:
            raise EngineError("Bad SAVE error in HZ7 field")

        if f.read():
            raise EngineError("Bad SAVE length (extra data)")



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# romiq.kh@gmail.com, 2015

import sys, os
import traceback
import tkinter

from tkguibrowser import TkBrowser, hlesc, cesc, fmt_hl, fmt_hl_len, fmt_arg, \
    fmt_dec, fmt_dec_len

APPNAME = "Test tkinter gui browser"
VERSION = "v0.1 2015-06-20"

class App(TkBrowser):
    def __init__(self, master):
        super().__init__(master)

    def init_gui(self):
        self.master.title(APPNAME)
        # path
        if hasattr(sys, 'frozen'):
            self.app_path = sys.executable
        else:
            self.app_path = __file__
        self.app_path = os.path.abspath(os.path.dirname(self.app_path))

    def create_widgets(self):
        super().create_widgets()

        self.path_handler["test"] = [self.path_test, "Tests"]

        repath = "/"
        for cmd, arg in self.start_act:
            if cmd == "open":
                repath = ""
                if not self.open_path(arg):
                    print("DEBUG: stop opening after " + arg)
                    repath = ""
                    break
        if repath:
            self.open_path(repath)


    def create_menu(self):
        super().create_menu()
        self.menufile = tkinter.Menu(self.master, tearoff = 0)
        self.menubar.add_cascade(menu = self.menufile,
                label = "File")
        self.menufile.add_separator()
        self.menufile.add_command(
                command = self.on_exit,
                label = "Quit")

    def path_default(self, path):
        self.switch_view(0)
        self.update_gui("Outline")
        self.clear_info()
        self.add_info("Open path\n\n" + str(path))

        self.add_info("\n\n")
        self.add_info("<a href=\"http://example.com/test\">example.com</a>\n")
        self.add_info("\n\n<b>This is multi-")
        self.add_info("line\nbold</b> text\n")
        self.add_info("\n\n<u>This is multi-")
        self.add_info("line\nunderline</u> text\n")
        self.add_info("\n\n<font color=\"#ff00FF\">This is multi-")
        self.add_info("line\ncolor</font> text\n")

        self.add_info("\n<font bg=\"red\" color=\"white\"> <b>White</b> on <i>red</i> </font>\n")

        self.insert_lb_act("Testing", "/test")
        return True

    def path_test(self, path):
        if len(path) > 1:
            return self.path_test_item(path)
        if self.last_path != ("test",):
            self.update_gui("Test")
            self.insert_lb_act("Outline", [])
            self.insert_lb_act("-", None)
            self.insert_lb_act("Info", "/test/info")
            self.insert_lb_act("Image", "/test/image")
            self.switch_view(0)
            self.clear_info()
            self.add_info("Select test from outline")
        return True

    def path_test_item(self, path):
        def display_page():
            item = None
            path = self.curr_path
            if len(path) > 2:
                item = path[2]
            self.clear_info()
            sm = self.gl_state.get("test.info.mode", 0)
            if item is None:
                sm = -1
            self.upd_toolgrp(self.curr_state["gbtns"], sm)
            if item is None:
                self.switch_view(0)
                self.add_info("Select item " + path[1])
            else:
                if path[1] == "image":
                    self.switch_view(1)
                    self.main_image = tkinter.PhotoImage(file = "img/splash.gif")
                elif path[1] == "info":
                    self.switch_view(0)
                    self.add_info("Information panel for {}\n".format(path))
                    self.add_info("Local mode {}\n".format(
                        self.curr_state.get("mode", None)))
                    self.add_info("Global mode {}\n".format(
                        self.gl_state.get("test.info.mode", None)))
                    for i in range(100):
                        self.add_info("  Item {}\n".format(i))

        if self.last_path[:2] != ("test", path[1]):
            self.update_gui("Test %s" % path[1])
            self.insert_lb_act("Testing", "/test")
            self.insert_lb_act("-", None)
            for i in range(15):
                self.insert_lb_act("{} #{}".format(path[1], i),
                    path[:2] + (i,), i)
            # create mode buttons
            def sw_mode1():
                print("Mode 1")
                self.curr_state["mode"] = 1
                self.curr_state["btn1"].config(state = tkinter.DISABLED)
                self.curr_state["btn2"].config(state = tkinter.NORMAL)
                display_page()
            def sw_mode2():
                print("Mode 2")
                self.curr_state["mode"] = 2
                self.curr_state["btn1"].config(state = tkinter.NORMAL)
                self.curr_state["btn2"].config(state = tkinter.DISABLED)
                display_page()
            self.curr_state["btn1"] = self.add_toolbtn("Mode 1", sw_mode1)
            self.curr_state["btn1"].config(state = tkinter.DISABLED)
            self.curr_state["btn2"] = self.add_toolbtn("Mode 2", sw_mode2)
            # we store buttons in local state
            self.curr_state["gbtns"] = self.add_toolgrp(None, "test.info.mode",
                {0: "mode 1", 1: "mode 2", 2: "mode 3"}, display_page)

        # change
        item = None
        if len(path) > 2:
            # index
            self.select_lb_item(path[2])
        else:
            self.select_lb_item(None)
        # display
        display_page()
        return True

def main():
    root = tkinter.Tk()
    app = App(master = root)
    argv = sys.argv[1:]
    while len(argv) > 0:
        if argv[0] == "--nooutline": # open data
            app.gui_setup["outline"] = False
            argv = argv[1:]
        else:
            app.start_act.append(["open", argv[0]])
            argv = argv[1:]
    app.mainloop()

if __name__ == "__main__":
    main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# romiq.kh@gmail.com, 2015

import math
import traceback
from html.parser import HTMLParser

import tkinter
from tkinter import ttk, font, filedialog, messagebox
from idlelib.redirector import WidgetRedirector

# Image processing
try:
    from PIL import Image
except ImportError:
    Image = None

try:
    from PIL import ImageTk
except ImportError:
    ImageTk = None

def hlesc(value):
    if value is None:
        return "None"
    return value.replace("\\", "\\\\").replace("<", "\\<").replace(">", "\\>")

def cesc(value):
    return value.replace("\\", "\\\\").replace("\"", "\\\"")

def fmt_hl(loc, desc):
    return "<a href=\"{}\">{}</a>".format(loc, desc)

def fmt_hl_len(loc, desc, ln):
    sz = max(ln - len(desc), 0)
    return " "*sz + fmt_hl(loc, desc)

def fmt_arg(value):
    if value < 10:
        return "{}".format(value)
    elif value == 0xffff:
        return "-1"
    else:
        return "0x{:X}".format(value)

def fmt_dec(value, add = 0):
    return "{{:{}}}".format(fmt_dec_len(value, add))

def fmt_dec_len(value, add = 0):
    if value == 0:
        d = 1
    else:
        d = int(math.log10(value)) + 1
    d += add
    return d

# thanx to http://effbot.org/zone/tkinter-text-hyperlink.htm
class HyperlinkManager(HTMLParser):

    def __init__(self, text):
        self.text = text
        self.text.tag_config("hyper", foreground = "blue", underline = 1)
        self.text.tag_bind("hyper", "<Enter>", self._enter)
        self.text.tag_bind("hyper", "<Leave>", self._leave)
        self.text.tag_bind("hyper", "<Button-1>", self._click)
        bold_font = font.Font(text, self.text.cget("font"))
        bold_font.configure(weight = "bold")
        self.text.tag_config("bold", font = bold_font)
        italic_font = font.Font(text, self.text.cget("font"))
        italic_font.configure(slant = "italic")
        self.text.tag_config("italic", font = italic_font)
        self.text.tag_config("underline", underline = 1)
        self.parser = HTMLParser()
        self.parser.handle_starttag = self.handle_starttag
        self.parser.handle_endtag = self.handle_endtag
        self.parser.handle_data = self.handle_data
        self.reset()

    def reset(self):
    	self.links = {}
    	self.colors = []
    	self.bgs = []
    	self.colorbgs = []

    def add(self, action):
        # add an action to the manager.  returns tags to use in
        # associated text widget
        tag = "hyper-{}".format(len(self.links))
        self.links[tag] = action
        return "hyper", tag

    def color(self, color):
        tag = "color-{}".format(color)
        if tag not in self.colors:
            self.colors.append(tag)
            self.text.tag_config(tag, foreground = color)
            self.text.tag_raise("hyper")
        return (tag,)

    def bg(self, color):
        tag = "bg-{}".format(color)
        if tag not in self.bgs:
            self.bgs.append(tag)
            self.text.tag_config(tag, background = color)
            self.text.tag_raise("hyper")
        return (tag,)

    def colorbg(self, color, bg):
        tag = "colorbg-{}|{}".format(color, bg)
        if tag not in self.colorbgs:
            self.colorbgs.append(tag)
            self.text.tag_config(tag, foreground = color, background = bg)
            self.text.tag_raise("hyper")
        return (tag,)

    def _enter(self, event):
        self.text.config(cursor = "hand2")

    def _leave(self, event):
        self.text.config(cursor = "")

    def _click(self, event):
        for tag in self.text.tag_names(tkinter.CURRENT):
            if tag[:6] == "hyper-":
                self.links[tag]()
                return

    def handle_starttag(self, tag, attrs):
        tagmap = {"b": "bold", "i": "italic", "u": "underline"}
        if tag in tagmap:
            self.parser_tags.append([tagmap[tag]])
        elif tag == "a":
            ref = ""
            for k, v in attrs:
                if k == "href":
                    ref = v
            self.parser_tags.append(self.add(self.parser_handler(ref)))
        elif tag == "font":
            color = ""
            bg = ""
            for k, v in attrs:
                if k == "color":
                    color = v
                elif k == "bg":
                    bg = v
            if color and bg:
                self.parser_tags.append(self.colorbg(color, bg))
            elif bg:
                self.parser_tags.append(self.bg(bg))
            else:
                self.parser_tags.append(self.color(color))

    def handle_endtag(self, tag):
        self.parser_tags = self.parser_tags[:-1]

    def handle_data(self, data):
        self.parser_widget.insert(tkinter.INSERT, data, \
            tuple(reversed([x for x in self.parser_tags for x in x])))

    def add_markup(self, text, widget, handler):
        self.parser_tags = []
        self.parser_widget = widget
        self.parser_handler = handler
        self.parser.reset()
        self.parser.feed(text)
        return


# thanx http://tkinter.unpythonic.net/wiki/ReadOnlyText
class ReadOnlyText(tkinter.Text):

    def __init__(self, *args, **kwargs):
        tkinter.Text.__init__(self, *args, **kwargs)
        self.redirector = WidgetRedirector(self)
        self.insert = \
            self.redirector.register("insert", lambda *args, **kw: "break")
        self.delete = \
            self.redirector.register("delete", lambda *args, **kw: "break")


class TkBrowser(tkinter.Frame):

    def __init__(self, master):
        tkinter.Frame.__init__(self, master)
        self.pack(fill = tkinter.BOTH, expand = 1)
        self.pad = None

        # gui
        self.path_handler = {}
        self.curr_main = -1 # 0 - frame, 1 - canvas
        self.curr_path = []
        self.curr_help = ""
        self.last_path = [None]
        self.curr_gui = []
        self.curr_state = {} # local state for location group
        self.curr_markup = "" # current unparsed markup data (unclosed tags, etc)
        self.curr_lb_acts = None
        self.curr_lb_idx = None
        self.hist = []
        self.histf = []
        self.gl_state = {} # global state until program exit
        self.start_act = []
        self.init_gui() # init custom gui data

        # canvas
        self.need_update = False
        self.canv_view_fact = 1
        self.main_image = tkinter.PhotoImage(width = 1, height = 1)
        # add on_load handler
        self.after_idle(self.on_first_display)

    def init_gui(self):
        pass

    def update_after(self):
        if not self.need_update:
            self.after_idle(self.on_idle)
            self.need_update = True

    def on_idle(self):
        self.need_update = False
        self.update_canvas()

    def on_first_display(self):
        fnt = font.Font()
        try:
            self.pad = fnt.measure(":")
        except:
            self.pad = 5
        self.create_widgets()
        self.create_menu()

    def on_help(self):
        pass

    def on_back(self):
        if len(self.hist) > 1:
            np = self.hist[-2:-1][0]
            self.histf = self.hist[-1:] + self.histf
            self.hist = self.hist[:-1]
            self.open_path(np[0], False)

    def on_forward(self):
        if len(self.histf) > 0:
            np = self.histf[0]
            self.histf = self.histf[1:]
            self.hist.append(np)
            self.open_path(np[0], False)

    def create_widgets(self):
        ttk.Style().configure("Tool.TButton", width = -1) # minimal width
        ttk.Style().configure("TLabel", padding = self.pad)
        ttk.Style().configure('Info.TFrame', background = 'white', \
            foreground = "black")

        # toolbar
        self.toolbar = ttk.Frame(self)
        self.toolbar.pack(fill = tkinter.BOTH)
        btns = [
            ["Outline", lambda: self.open_path("")],
            ["Help", self.on_help],
            [None, None],
            ["<-", self.on_back],
            ["->", self.on_forward],
        ]
        for text, cmd in btns:
            if text is None:
                frm = ttk.Frame(self.toolbar, width = self.pad,
                    height = self.pad)
                frm.pack(side = tkinter.LEFT)
                continue
            btn = ttk.Button(self.toolbar, text = text, \
                style = "Tool.TButton", command = cmd)
            btn.pack(side = tkinter.LEFT)
        frm = ttk.Frame(self.toolbar, width = self.pad, height = self.pad)
        frm.pack(side = tkinter.LEFT)

        # main panel
        self.pan_main = ttk.PanedWindow(self, orient = tkinter.HORIZONTAL)
        self.pan_main.pack(fill = tkinter.BOTH, expand = 1)

        # leftpanel
        self.frm_left = ttk.Frame(self.pan_main)
        self.pan_main.add(self.frm_left)
        # main view
        self.frm_view = ttk.Frame(self.pan_main)
        self.pan_main.add(self.frm_view)
        self.frm_view.grid_rowconfigure(0, weight = 1)
        self.frm_view.grid_columnconfigure(0, weight = 1)
        self.scr_view_x = ttk.Scrollbar(self.frm_view,
            orient = tkinter.HORIZONTAL)
        self.scr_view_x.grid(row = 1, column = 0, \
            sticky = tkinter.E + tkinter.W)
        self.scr_view_y = ttk.Scrollbar(self.frm_view)
        self.scr_view_y.grid(row = 0, column = 1, sticky = \
            tkinter.N + tkinter.S)
        # canvas
        self.canv_view = tkinter.Canvas(self.frm_view, height = 150,
            bd = 0, highlightthickness = 0,
            scrollregion = (0, 0, 50, 50),
            )
        # don't forget
        #   canvas.config(scrollregion=(left, top, right, bottom))
        self.canv_view.bind('<Configure>', self.on_resize_view)
        self.canv_view.bind('<ButtonPress-1>', self.on_mouse_view)

        # text
        self.text_view = ReadOnlyText(self.frm_view,
            highlightthickness = 0,
            )
        self.text_hl = HyperlinkManager(self.text_view)
        self.text_view.bind('<Configure>', self.on_resize_view)

    def create_menu(self):
        self.menubar = tkinter.Menu(self.master)
        self.master.configure(menu = self.menubar)

    def on_exit(self):
        self.master.destroy()

    def on_mouse_view(self, event):
        self.update_after()

    def on_resize_view(self, event):
        self.update_after()

    def parse_path(self, loc):
        if isinstance(loc, str):
            path = []
            if loc[:1] == "/":
                loc = loc[1:]
            if loc != "":
                for item in loc.split("/"):
                    try:
                        path.append(int(item, 10))
                    except:
                        path.append(item)
        else:
            path = loc
        path = tuple(path)
        while path[-1:] == ("",):
            path = path[:-1]
        return path

    def desc_path(self, loc):
        path = self.parse_path(loc)
        if len(path) > 0:
            if path[0] in self.path_handler:
                desc = self.path_handler[path[0]][1]
                if callable(desc):
                    return desc(path)
                elif desc:
                    return desc
        return self.desc_default(path)

    def update_canvas(self):
        if self.curr_main == 0:
            return
        # draw grahics
        c = self.canv_view
        c.delete(tkinter.ALL)

        w = self.canv_view.winfo_width()
        h = self.canv_view.winfo_height()
        if (w == 0) or (h == 0):
            return

        scale = 0

        # Preview image
        if not isinstance(self.main_image, tkinter.PhotoImage):
            mw, mh = self.main_image.size
            if scale == 0: # Fit
                try:
                    psc = w / h
                    isc = mw / mh
                    if psc < isc:
                        fact = w / mw
                    else:
                        fact = h / mh
                except:
                    fact = 1.0
            else:
                fact = scale
            pw = int(mw * fact)
            ph = int(mh * fact)
            img = self.main_image.resize((pw, ph), Image.ANTIALIAS)
            self.canv_image = ImageTk.PhotoImage(img)
        else:
            mw = self.main_image.width()
            mh = self.main_image.height()
            if scale == 0: # Fit
                try:
                    psc = w / h
                    isc = mw / mh
                    if psc < isc:
                        if w > mw:
                            fact = w // mw
                        else:
                            fact = -mw // w
                    else:
                        if h > mh:
                            fact = h // mh
                        else:
                            fact = -mh // h
                except:
                    fact = 1
            else:
                fact = scale
            self.canv_image = self.main_image.copy()
            if fact > 0:
                self.canv_image = self.canv_image.zoom(fact)
            else:
                self.canv_image = self.canv_image.subsample(-fact)
            self.canv_image_fact = fact

            # place on canvas
            if fact > 0:
                pw = mw * fact
                ph = mh * fact
            else:
                pw = mw // -fact
                ph = mh // -fact

        cw = max(pw, w)
        ch = max(ph, h)
        c.config(scrollregion = (0, 0, cw - 2, ch - 2))
        #print("Place c %d %d, p %d %d" % (cw, ch, w, h))
        c.create_image(cw // 2, ch // 2, image = self.canv_image)

    def make_image(self, imgobj):
        if imgobj.image is not None:
            return imgobj.image
        width = imgobj.width
        height = imgobj.height
        data = imgobj.rgb
        # create P6
        phdr = ("P6\n{} {}\n255\n".format(width, height))
        rawlen = width * height * 3 # RGB
        #phdr = ("P5\n{} {}\n255\n".format(width, height))
        #rawlen = width * height
        phdr = phdr.encode("UTF-8")

        if len(data) > rawlen:
            # truncate
            pdata = data[:rawlen]
        if len(data) < rawlen:
            # fill gap
            gap = bytearray()
            data += b"\xff" * (rawlen - len(data))
        p = bytearray(phdr)
        # fix UTF-8 issue
        for ch in data:
            if ch > 0x7f:
                p += bytes((0b11000000 |\
                    ch >> 6, 0b10000000 |\
                    (ch & 0b00111111)))
            else:
                p += bytes((ch,))
        image = tkinter.PhotoImage(width = width, height = height, \
            data = bytes(p))
        return image

    def update_gui(self, text = "<Undefined>"):
        self.last_path = self.curr_path
        # cleanup
        for item in self.curr_gui:
            item()
        self.curr_gui = []
        self.curr_state = {} # save state across moves
        # left listbox
        lab = tkinter.Label(self.frm_left, text = text)
        lab.pack()
        frm_lb = ttk.Frame(self.frm_left)
        frm_lb.pack(fill = tkinter.BOTH, expand = 1)
        frm_lb.grid_rowconfigure(0, weight = 1)
        frm_lb.grid_columnconfigure(0, weight = 1)
        scr_lb_x = ttk.Scrollbar(frm_lb, orient = tkinter.HORIZONTAL)
        scr_lb_x.grid(row = 1, column = 0, sticky = tkinter.E + tkinter.W)
        scr_lb_y = ttk.Scrollbar(frm_lb)
        scr_lb_y.grid(row = 0, column = 1, sticky = tkinter.N + tkinter.S)
        frmlbpad = ttk.Frame(frm_lb, borderwidth = self.pad)
        lb = tkinter.Listbox(frm_lb,
            highlightthickness = 0,
            xscrollcommand = scr_lb_x.set,
            yscrollcommand = scr_lb_y.set)
        lb.grid(row = 0, column = 0, \
            sticky = tkinter.N + tkinter.S + tkinter.E + tkinter.W)
        scr_lb_x.config(command = lb.xview)
        scr_lb_y.config(command = lb.yview)
        self.curr_gui.append(lambda:lb.grid_remove())
        self.curr_gui.append(lambda:lab.pack_forget())
        self.curr_gui.append(lambda:frm_lb.pack_forget())
        lb.bind("<Double-Button-1>", self.on_left_listbox)
        lb.bind("<Return>", self.on_left_listbox)
        # actions on listbox
        self.curr_lb = lb
        self.curr_lb_acts = []
        self.curr_lb_idx = {}

    def switch_view(self, main):
        # main view
        if main == self.curr_main: return
        last = self.curr_main
        self.curr_main = main
        rw = None
        rh = None
        if main == 0:
            self.canv_view.delete(tkinter.ALL)
            self.canv_view.grid_forget()
            self.text_view.grid(row = 0, column = 0, \
                sticky = tkinter.N + tkinter.S + tkinter.E + tkinter.W)
            self.text_view.configure(
                xscrollcommand = self.scr_view_x.set,
                yscrollcommand = self.scr_view_y.set
            )
            self.scr_view_x.config(command = self.text_view.xview)
            self.scr_view_y.config(command = self.text_view.yview)
        else:
            if last == 0:
                rw = self.text_view.winfo_width()
                rh = self.text_view.winfo_height()
            self.canv_view.delete(tkinter.ALL)
            self.text_view.grid_forget()
            self.canv_view.grid(row = 0, column = 0, \
                sticky = tkinter.N + tkinter.S + tkinter.E + tkinter.W)
            self.canv_view.configure(
                xscrollcommand = self.scr_view_x.set,
                yscrollcommand = self.scr_view_y.set
            )
            self.scr_view_x.config(command = self.canv_view.xview)
            self.scr_view_y.config(command = self.canv_view.yview)
            if rh:
                print(rh)
                self.canv_view.height = rh
                print(self.canv_view.winfo_height())

    def clear_info(self):
        self.text_view.delete(0.0, tkinter.END)

    def add_text(self, text):
        self.end_markup()
        self.text_view.insert(tkinter.INSERT, text)

    def add_info(self, text):
        self.curr_markup += text

    def end_markup(self):
        if not self.curr_markup: return
        def make_cb(path):
            def cb():
                if path[:5] == "http:" or path[:6] == "https:":
                    return self.open_http(path)
                return self.open_path(path)
            return cb
        self.text_hl.add_markup(self.curr_markup, self.text_view, make_cb)
        self.curr_markup = ""

    def insert_lb_act(self, name, act, key = None):
        if key is not None:
            self.curr_lb_idx[key] = len(self.curr_lb_acts)
        self.curr_lb_acts.append((name, act))
        if name == "-" and act is None:
            self.curr_lb.insert(tkinter.END, "")
        else:
            self.curr_lb.insert(tkinter.END, " " + name)

    def select_lb_item(self, key):
        idx = self.curr_lb_idx.get(key, None)
        need = (idx is not None)
        idxs = "{}".format(idx)
        for sel in self.curr_lb.curselection():
            if sel == idxs:
                need = False
            else:
                self.curr_lb.selection_clear(sel)
        if need:
            self.curr_lb.selection_set(idxs)
        if idx is not None:
            self.curr_lb.see(idxs)

    def on_left_listbox(self, event):
        def currsel():
            try:
                num = self.curr_lb.curselection()[0]
                num = int(num)
            except:
                return None
            return num

        if self.curr_lb_acts:
            act = self.curr_lb_acts[currsel()]
            if act[1] is not None:
                self.open_path(act[1])

    def add_toolbtn(self, text, cmd):
        if text is None:
            frm = ttk.Frame(self.toolbar, width = self.pad, height = self.pad)
            frm.pack(side = tkinter.LEFT)
            self.curr_gui.append(lambda:frm.pack_forget())
            return
        btn = ttk.Button(self.toolbar, text = text, \
            style = "Tool.TButton", command = cmd)
        btn.pack(side = tkinter.LEFT)
        self.curr_gui.append(lambda:btn.pack_forget())
        return btn

    def add_toollabel(self, text):
        lab = ttk.Label(self.toolbar, text = text)
        lab.pack(side = tkinter.LEFT)
        self.curr_gui.append(lambda:lab.pack_forget())
        return lab

    def add_toolgrp(self, label, glkey, items, cbupd):
        def makecb(v, g):
            def btncb():
                self.gl_state[g] = v
                cbupd()
            return btncb
        if label:
            self.add_toollabel(label)
        kl = list(items.keys())
        kl.sort()
        res = []
        for k in kl:
            b = self.add_toolbtn(items[k], makecb(k, glkey))
            res.append([b, k])
        return res

    def upd_toolgrp(self, btns, state):
        for btn, idx in btns:
            if idx != state and state != -1:
                btn.config(state = tkinter.NORMAL)
            else:
                btn.config(state = tkinter.DISABLED)

    def clear_hist(self):
        self.hist = self.hist[-1:]
        self.histf = []

    def open_http(self, path):
        messagebox.showinfo(parent = self, title = "URL", message = path)

    def open_path(self, loc, withhist = True):
        path = self.parse_path(loc)
        if withhist:
            self.hist.append([path])
            self.histf = []
        print("DEBUG: Open", path)
        self.curr_path = path
        if len(path) > 0:
            self.curr_help = path[0]
        else:
            self.curr_help = ""
        try:
            if len(path) > 0 and path[0] in self.path_handler:
                res = self.path_handler[path[0]][0](path)
            else:
                res = self.path_default(path)
        except Exception:
            self.switch_view(0)
            self.add_text("\n" + "="*20 + "\n" + traceback.format_exc())
            res = True
        self.end_markup()
        return res

    def path_default(self, path):
        self.switch_view(0)
        self.clear_info()
        self.add_info("Open path\n\n" + str(path))
        return True



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# romiq.kh@gmail.com, 2014

import os, sys
import traceback
import argparse
import io
import hashlib

import petka
import petka.engine

APPNAME = "P1&2 Compiler and decompiler"
VERSION = "v0.3h 2015-01-12"

class ScriptSyntaxError(Exception): pass

def find_in_folder(folder, name, ifnot = True):
    for item in os.listdir(folder):
        if item.upper() == name.upper():
            return os.path.join(folder, item)
    if ifnot:
        return os.path.join(folder, name)
    else:
        return None

class P12Compiler:
    def __init__(self):
        pass

    # =======================================================================
    # compiler utils
    # =======================================================================

    def tokenizer(self, source, enc):
        for lineno, line in enumerate(source.readlines(), 1):
            line = line.decode(enc or "UTF-8").strip()
            # eliminate comment
            nline = []
            nmode = 0 # 0 - common, 2 - in string
            nitem = ""
            nitemesc = False
            for ch in line:
                if nmode == 1:
                    if nitemesc:
                        nitem += ch
                        nitemesc = False
                    else:
                        if ch == "\\":
                            # esc-sequence
                            nitemesc = True
                            continue
                        elif ch == "\"":
                            nitem += ch
                            nmode = 0
                            if nitem[:1] == nitem[-1:] == "\"":
                                nitem = nitem[1:-1]
                            nline.append(nitem)
                            nitem = ""
                            nitemesc = False
                            continue
                        nitem += ch
                else:
                    if ch == "\"":
                        nmode = 1
                    elif ch == "#":
                        break
                    elif ch == " " or ch == "\t":
                        if len(nitem) > 0:
                            nline.append(nitem)
                            nitem = ""
                        continue
                    nitem += ch

            if len(nitem) > 0:
                nline.append(nitem)

            yield lineno, nline


    def check8(self, value, name, lineno):
        if value == -1:
            return 0xff
        if value < 0:
            raise ScriptSyntaxError("Error at {}: value for \"{}\" too "\
                "small - {}".format(lineno, name, value))
        if value > 0xff:
            raise ScriptSyntaxError("Error at {}: value for \"{}\" too "\
                "big - {} > 0xff".format(lineno, name, value))
        return value

    def check16(self, value, name, lineno):
        if value == -1:
            return 0xffff
        if value < 0:
            raise ScriptSyntaxError("Error at {}: value for \"{}\" too "\
                "small - {}".format(lineno, name, value))
        if value > 0xffff:
            raise ScriptSyntaxError("Error at {}: value for \"{}\" too "\
                "big - {} > 0xffff".format(lineno, name, value))
        return value

    def check32(self, value, name, lineno):
        if value == -1:
            return 0xffffffff
        if value < 0:
            raise ScriptSyntaxError("Error at {}: value for \"{}\" too "\
                "small - {}".format(lineno, name, value))
        if value > 0xffffffff:
            raise ScriptSyntaxError("Error at {}: value for \"{}\" too "\
                "big - {} > 0xffffffff".format(lineno, name, value))
        return value

    def convertnum(self, value):
        num = None
        if value[:2].upper() == "0X":
            try:
                num = int(value[2:], 16)
            except:
                pass
        else:
            try:
                num = int(value, 10)
            except:
                pass
        return num

    def checkusedid(self, ident, lineno):
        if ident in self.usedid:
            raise ScriptSyntaxError("Error at {}: identificator \"{}\" "\
                "already used at line {}".format(lineno, ident, \
                self.usedid[ident][0]))
        self.usedid[ident] = [lineno, None]

    def setidentvalue(self, ident, value):
        if self.usedid[ident][1] is None:
            self.usedid[ident][1] = value
        else:
            raise ScriptSyntaxError("Redefine value for \"{}\" ({}) from "\
                "\"{}\" to \"{}\"".format(ident, self.usedid[ident][0], \
                self.usedid[ident][0], value))

    def getidentvalue(self, ident):
        return self.usedid[ident][1]

    def checkident(self, ident, name, lineno):
        if ident.upper() in self.reservedid:
            raise ScriptSyntaxError("Error at {}: identificator \"{}\" "\
                "in {} forbidden".format(lineno, ident, name))
        if ident not in self.usedid:
            raise ScriptSyntaxError("Error at {}: identificator \"{}\" "\
                "in {} not found".format(lineno, ident, name))
        if self.usedid[ident][1] is None:
            raise ScriptSyntaxError("Error at {}: identificator \"{}\" "\
                "in {} has no value (internal error)".\
                format(lineno, ident, name))

    def checkstruct(self, name, struct, data, lineno):
        try:
            struct.build(data)
        except Exception as e:
            msg = "Internal error {}".format(name)
            if lineno is not None:
                msg += ", at line {}".format(lineno)
            raise ScriptSyntaxError(msg + "\n" + str(e))

    def convertargs(self, fmt, args, lineno):
        # fmt = [(name1, check1, withident),]
        pargs = []
        argnum = [None] * len(args)
        for i, arg in enumerate(args):
            argnum[i] = self.convertnum(arg)
            if argnum[i] is None:
                if not fmt[i][2]:
                    raise ScriptSyntaxError("Error at {}: bad number for {}".\
                        format(lineno, fmt[i][0]))
                self.checkident(arg, fmt[i][0], lineno)
                argnum[i] = self.getidentvalue(arg)
            argnum[i] = fmt[i][1](argnum[i], fmt[i][0], lineno)
        return argnum

    # =======================================================================
    # compile SCRIPT.DAT
    # =======================================================================

    def compile_script(self, source, destfolder, enc = None, \
            st_scr = None, st_bkg = None, st_res = None):

        pe = petka.Engine()
        pe.init_empty("cp1251")

        mode = 0 # 0 - common, 1 - object/scene, 2 - on
        mode1tp = None

        # used identificators
        self.usedid = {}
        # compiled resources
        compres = []
        # compiled obj
        compobj = []
        # compled scenes
        compscene = []
        # current item
        compitem = None
        # obj/scene used numbers
        compitemused = {}
        # current action
        compact = None

        revOPS = {}
        for ok, ov in petka.OPCODES.items():
            revOPS[ov[0]] = ok
        self.reservedid = ["THIS", "OBJ", "ENDOBJ", "SCENE", "ENDSCENE", \
            "RES", "REF", "ON", "ENDON"] + list(revOPS.keys())

        for lineno, tokens in self.tokenizer(source, enc):
            if len(tokens) == 0:
                continue
            if mode == 0:
                # accept: RES, OBJ, SCENE
                cmd = tokens[0].upper()
                if cmd == "RES":
                    if len(tokens) != 4:
                        raise ScriptSyntaxError("Error at {}: RES syntax "\
                            "error".format(lineno, cmd))
                    self.checkusedid(tokens[1], lineno)
                    compres.append((lineno, tokens[1], \
                        tokens[2], tokens[3]))
                elif cmd == "OBJ" or cmd == "SCENE":
                    # object / scene
                    mode = 1
                    mode1tp = cmd
                    # check syntax
                    if len(tokens) != 4:
                        raise ScriptSyntaxError("Error at {}: {} syntax "\
                            "error".format(lineno, cmd))
                    self.checkusedid(tokens[1], lineno)
                    num = self.convertnum(tokens[2])
                    if num is not None:
                        if num in compitemused:
                            raise ScriptSyntaxError("Error at {}: {} "\
                                "number {} already used at {}".format(\
                                    lineno, cmd.lower(), num, \
                                    compitemused[num]))
                        else:
                            compitemused[num] = lineno
                        num = self.check16(num, "{} number".format(cmd.lower()),
                            lineno)
                        self.setidentvalue(tokens[1], num)
                    else:
                        raise ScriptSyntaxError("Error at {}: {} number bad "\
                            "format \"{}\" ".format(lineno, cmd.lower(),
                                tokens[2]))
                    compitem = {"tp": cmd, "lineno": lineno, \
                        "ident": tokens[1], "num": num, \
                        "name": tokens[3], "ref": None, "acts": []}
                else:
                    raise ScriptSyntaxError("Error at {}: unknown syntax "\
                        "\"{}\"".format(lineno, cmd))
            elif mode == 1:
                # accept: REF, ON, ENDOBJ, ENDSCENE
                cmd = tokens[0].upper()
                if cmd == "REF" and mode1tp == "SCENE":
                    if compitem["ref"] is None:
                        compitem["ref"] = []
                    # check format
                    if len(tokens) < 4 or len(tokens) > 7:
                        raise ScriptSyntaxError("Error at {}: unknown REF "\
                            "syntax in {}".format(lineno, mode1tp))
                    while len(tokens) < 7:
                        tokens.append("-1")
                    compitem["ref"].append([lineno] + tokens[1:])
                elif cmd == "ZEROREF" and mode1tp == "SCENE":
                    # zero rererence at all
                    if compitem["ref"] is not None:
                        # can't empty any
                        raise ScriptSyntaxError("Error at {}: ref already "\
                            "specified".format(lineno))
                    compitem["ref"] = []
                elif cmd == "ON":
                    # on action
                    if len(tokens) != 2 and len(tokens) != 4:
                        raise ScriptSyntaxError("Error at {}: unknown ON "\
                            "syntax in {}".format(lineno, mode1tp))
                    son = self.convertnum(tokens[1])
                    if son is None:
                        if tokens[1].upper() in revOPS:
                            son = revOPS[tokens[1].upper()]
                        else:
                            raise ScriptSyntaxError("Error at {}: unknown ON "\
                                "OPREF \"{}\" in {}".\
                                    format(lineno, tokens[1], mode1tp))
                    else:
                        son = self.check16(son, "ON opref", lineno)
                    if len(tokens) == 2:
                        tokens += ["-1", "-1"]
                    status = self.convertnum(tokens[2])
                    if status is None:
                        raise ScriptSyntaxError("Error at {}: unknown ON "\
                            "status \"{}\" in {}".\
                                format(lineno, tokens[2], mode1tp))
                    status = self.check8(status, "ON status", lineno)
                    compact = {"son": son, "ops": [], "status": status, \
                        "sonref": tokens[3], "lineno": lineno}
                    mode = 2
                elif cmd == "END" + mode1tp and len(tokens) == 1:
                    if mode1tp == "OBJ":
                        compobj.append(compitem)
                    else:
                        compscene.append(compitem)
                    compitem = None
                    # return
                    mode = 0
                else:
                    raise ScriptSyntaxError("Error at {}: unknown syntax "\
                        "\"{}\" in {}".\
                        format(lineno, cmd, mode1tp))
            elif mode == 2:
                # accept: ENDON, OPCODE, any number
                cmd = tokens[0].upper()
                if cmd == "ENDON" and len(tokens) == 1:
                    compitem["acts"].append(compact)
                    compact = None
                    mode = 1
                else:
                    # check format
                    if len(tokens) < 2 or len(tokens) > 5:
                        raise ScriptSyntaxError("Error at {}: unknown OP "\
                            "syntax in {}".format(lineno, mode1tp))
                    while len(tokens) < 5:
                        tokens.append("-1")
                    op = {}
                    if cmd in revOPS:
                        # opcode
                        opcode = revOPS[cmd]
                    else:
                        opcode = self.convertnum(cmd)
                        if opcode is None:
                            raise ScriptSyntaxError("Error at {}: unknown OP "\
                                "\"{}\" in ON".\
                                format(lineno, cmd))
                        opcode = self.check16(opcode, "opcode", lineno)
                    compact["ops"].append({"opcode": opcode, "lineno": lineno, \
                        "obj_ref": tokens[1], "args": tokens[2:]})
            else:
                raise ScriptSyntaxError("Error at {}: unknown parser mode {}".\
                    format(lineno, mode))

        # check unclosed objects
        if mode != 0:
            raise ScriptSyntaxError("Error at {}: unfinished structure".\
                format(lineno))

        # second stage - RES
        # 1. capture res with numbers
        resused = {}
        for lineno, ident, resnum, path in compres:
            num = self.convertnum(resnum)
            if num is not None:
                if num in resused:
                    raise ScriptSyntaxError("Error at {}: RES number {} "\
                        "already used at {}".format(lineno, num, resused[num]))
                else:
                    resused[num] = lineno
                num = self.check16(num, "RES id", lineno)
                self.setidentvalue(ident, num)
            #elif resnum == "_":
            #    pass
            else:
                raise ScriptSyntaxError("Error at {}: RES number bad "\
                    "format \"{}\" ".format(lineno, resnum))
        # disabled, unsure about valid numbers range
        # 2. autonumber rest
        #autoresnum = 0xfffe
        #for lineno, ident, resnum, path in compres:
        #    if resnum == "_":
        #        while autoresnum > 0:
        #            num = autoresnum
        #            if num in resused:
        #                autoresnum -= 1
        #                continue
        #            else:
        #                self.setidentvalue(ident, num)
        #                resused[num] = lineno
        #                break
        #        if autoresnum <= 0:
        #            raise ScriptSyntaxError("Error at {}: out of resources "\
        #                "number".format(lineno))
        # 3. compile

        if destfolder is not None:
            f = open(os.path.join(destfolder, "resource.qrc"), "wb")
        else:
            f = st_res
        try:
            def writer(msg):
                f.write((msg + "\r\n").encode("CP1251"))
            writer("")
            for lineno, ident, resnum, path in compres:
                writer("{:d} == {}".format(self.getidentvalue(ident), path))
        finally:
            if destfolder is not None:
                f.close()
        print("RESOURCE.QRC saved: {} items".format(len(resused)))

        # second stage - OBJ
        def makerec(citem):
            item = petka.engine.ScrObject(citem["num"], citem["name"])
            item.acts = []
            for act in citem["acts"]:
                if act["sonref"].upper() == "THIS":
                    sonref = item.idx
                else:
                    sonref = self.convertnum(act["sonref"])
                if sonref is not None:
                    # direct number
                    sonref = self.check16(sonref, "ON object ref",
                        act["lineno"])
                else:
                    # get by ident
                    self.checkident(act["sonref"], "ON object ref",
                        act["lineno"])
                    sonref = self.getidentvalue(act["sonref"])
                onrec = petka.engine.ScrActObject(act["son"],
                    act["status"], sonref)
                onrec.ops = []
                for op in act["ops"]:
                    # object ref
                    if op["obj_ref"].upper() == "THIS":
                        opref = item.idx
                    else:
                        opref = self.convertnum(op["obj_ref"])
                    if opref is not None:
                        # direct number
                        opref = self.check16(opref, "OP object", op["lineno"])
                    else:
                        # get by ident
                        self.checkident(op["obj_ref"], "OP object",
                            op["lineno"])
                        opref = self.getidentvalue(op["obj_ref"])
                    # arguments
                    fmt = []
                    for i in range(3):
                        fmt.append(("OP argument {}".format(i + 1),
                            self.check16, True))
                    argnum = self.convertargs(fmt, op["args"], op["lineno"])
                    oprec =  petka.engine.ScrOpObject(opref, op["opcode"],
                        argnum[0], argnum[1], argnum[2])
                    onrec.ops.append(oprec)
                item.acts.append(onrec)
            return item

        for citem in compobj:
            objrec = makerec(citem)
            pe.objects.append(objrec)
            pe.obj_idx[objrec.idx] = objrec

        # second stage - SCENE
        backgrnd = []
        num_bkg = 0
        for citem in compscene:
            scenerec = makerec(citem)
            scenerec.refs = None
            pe.scenes.append(scenerec)

            if citem["ref"] is not None:
                scenerec.refs = []
                num_bkg += 1
                for ref in citem["ref"]:
                    fmt = [("REF object", self.check16, True)]
                    for i in range(5):
                        fmt.append(("REF argument {}".format(i + 1), \
                            self.check32, True))
                    argnum = self.convertargs(fmt, ref[1:], ref[0])
                    # build REF
                    if argnum[0] not in pe.obj_idx:
                        raise ScriptSyntaxError("Error at {}: referenced "
                            "object 0x{:x} not found".\
                            format(lineno, argnum[0]))
                    scenerec.refs.append([pe.obj_idx[argnum[0]], argnum[1],
                         argnum[2],  argnum[3],  argnum[4],  argnum[5]])


        if destfolder is not None:
            f = open(os.path.join(destfolder, "backgrnd.bg"), "wb")
        else:
            f = st_bkg
        try:
            pe.write_backgrnd(f)
        finally:
            if destfolder is not None:
                f.close()
        print("BACKGRND.BG saved: {} items".format(num_bkg))

        if destfolder is not None:
            f = open(os.path.join(destfolder, "script.dat"), "wb")
        else:
            f = st_scr
        try:
            pe.write_script(f)
        finally:
            if destfolder is not None:
                f.close()
        print("SCRIPT.DAT saved: {} objects, {} scenes".\
            format(len(pe.objects), len(pe.scenes)))


    # =======================================================================
    # compile DIALOGUE.FIX
    # =======================================================================
    def compile_dialog(self, source, destfolder, enc = None,
            st_fix = None, st_lod = None):

        pe = petka.Engine()
        pe.init_empty("cp1251")

        mode = 0 # 0 - common, 1 - grp, 2 - act, 3 - dlg
                 # 10 - msg (autoreset to 0)

        # used identificators
        self.usedid = {}

        revOPS = {}
        for ok, ov in petka.OPCODES.items():
            revOPS[ov[0]] = ok
        revDLGOPS = {}
        for ok, ov in petka.DLGOPS.items():
            revDLGOPS[ov[0]] = ok
        self.reservedid = ["MSG", "DLG", "DLGGRP", "ON", "ENDDLG",
            "ENDDLGGRP", "ENDON"] + list(revDLGOPS.keys())

        # msg array
        compmsg = []
        # current msg
        compmsgitem = None
        # grp array
        compgrp = []
        # current grp
        compgrpitem = None
        # current act
        compactitem = None
        # current dlg
        compdlgitem = None
        # dlgop count
        compdlgops = 0

        for lineno, tokens in self.tokenizer(source, enc):
            if len(tokens) == 0:
                continue
            if mode == 0:
                # accept: MSG, DLGGRP
                cmd = tokens[0].upper()
                if cmd == "MSG":
                    if len(tokens) < 3 or len(tokens) > 6:
                        raise ScriptSyntaxError("Error at {}: unknown MSG "\
                            "syntax".format(lineno))
                    while len(tokens) < 6:
                        tokens.append("0")
                    # check ident
                    self.checkusedid(tokens[1], lineno)
                    self.setidentvalue(tokens[1], len(compmsg))
                    # check wavfile name (1..12)
                    if len(tokens[2]) < 1 or len(tokens[2]) > 12:
                        raise ScriptSyntaxError("Error at {}: bad filename "\
                            "in MSG \"{}\"".format(lineno, tokens[2]))
                    mode = 10
                    compmsgitem = {"ident": tokens[1], "wav": tokens[2],
                        "args": tokens[3:], "lineno": lineno}
                    compmsg.append(compmsgitem)
                elif cmd == "DLGGRP":
                    # dlggrp
                    mode = 1
                    # check syntax
                    if len(tokens) < 2 or len(tokens) > 3:
                        raise ScriptSyntaxError("Error at {}: unknown DLGGRP "\
                            "syntax".format(lineno))
                    while len(tokens) < 3:
                        tokens.append("0")
                    compgrpitem = {"grp_id": tokens[1], "arg": tokens[2],
                        "acts": [], "lineno": lineno}
                else:
                    raise ScriptSyntaxError("Error at {}: unknown syntax "\
                        "\"{}\"".format(lineno, cmd))
            elif mode == 10:
                # MSG, 2nd string, one token - msg for subtitle
                if len(tokens) != 1:
                    raise ScriptSyntaxError("Error at {}: MSG syntax error, "\
                        "message required".format(lineno))
                compmsgitem["msg"] = tokens[0]
                compmsgitem = None
                mode = 0
            elif mode == 1:
                # accept: ON, ENDDLGGRP
                cmd = tokens[0].upper()
                if cmd == "ON":
                    # on
                    mode = 2
                    # check syntax
                    if len(tokens) < 3 or len(tokens) > 5:
                        raise ScriptSyntaxError("Error at {}: unknown ON "\
                            "syntax".format(lineno))
                    while len(tokens) < 5:
                        tokens.append("0")
                    if tokens[1] in revOPS:
                        # opcode
                        don = revOPS[tokens[1]]
                    else:
                        don = self.convertnum(tokens[1])
                        if don is None:
                            raise ScriptSyntaxError("Error at {}: unknown ON "\
                                "OPREF ""\"{}\" in ON".\
                                format(lineno, tokens[1]))
                        don = self.check16(don, "ON opref", lineno)
                    compactitem = {"don": don, "donref": tokens[2],
                        "args": tokens[3:], "dlgs": [], "lineno": lineno}
                elif cmd == "ENDDLGGRP" and len(tokens) == 1:
                    mode = 0
                    compgrp.append(compgrpitem)
                    compgrpitem = None
                else:
                    raise ScriptSyntaxError("Error at {}: unknown DLGGRP "\
                        "syntax \"{}\"".format(lineno, cmd))
            elif mode == 2:
                # accept: DLG, ENDON
                cmd = tokens[0].upper()
                if cmd == "DLG":
                    # dlg
                    mode = 3
                    # check syntax
                    if len(tokens) < 1 or len(tokens) > 3:
                        raise ScriptSyntaxError("Error at {}: unknown DLG "\
                            "syntax".format(lineno))
                    while len(tokens) < 3:
                        tokens.append("0")
                    compdlgitem = {"args": tokens[1:], "dlgops": [],
                        "lineno": lineno}
                elif cmd == "ENDON" and len(tokens) == 1:
                    mode = 1
                    compgrpitem["acts"].append(compactitem)
                    compactitem = None
                else:
                    raise ScriptSyntaxError("Error at {}: unknown ON "\
                        "syntax \"{}\"".format(lineno, cmd))
            elif mode == 3:
                # dlgopcode or ENDDLG
                cmd = tokens[0].upper()
                if cmd == "ENDDLG" and len(tokens) == 1:
                    compactitem["dlgs"].append(compdlgitem)
                    compdlgitem = None
                    mode = 2
                elif len(tokens) == 1 and cmd[-1:] == ":":
                    # label case
                    label = cmd[:-1]
                    self.checkusedid(tokens[0][:-1], lineno)
                    self.setidentvalue(tokens[0][:-1], compdlgops)
                else:
                    # check format
                    if len(tokens) < 1 or len(tokens) > 3:
                        raise ScriptSyntaxError("Error at {}: unknown DLGOP "\
                            "syntax in DLG".format(lineno))
                    while len(tokens) < 3:
                        tokens.append("0")
                    op = {}
                    if cmd in revDLGOPS:
                        # opcode
                        opcode = revDLGOPS[cmd]
                    else:
                        opcode = self.convertnum(cmd)
                        if opcode is None:
                            raise ScriptSyntaxError("Error at {}: unknown "\
                                "DLGOP \"{}\" in DLG".\
                                format(lineno, cmd))
                        opcode = self.check8(opcode, "dlgopcode", lineno)
                    compdlgops += 1
                    compdlgitem["dlgops"].append({"opcode": opcode,
                        "lineno": lineno, \
                        "msg_ref": tokens[2], "arg": tokens[1]})
            else:
                raise ScriptSyntaxError("Error at {}: unknown parser mode {}".\
                    format(lineno, mode))

        # check unclosed objects
        if mode != 0:
            raise ScriptSyntaxError("Error at {}: unfinished structure".\
                format(lineno))

        # second stage - MSG
        for msg in compmsg:
            # msg arguments
            fmt = []
            for i in range(3):
                fmt.append(("MSG argument {}".format(i + 1), \
                    self.check32, True))
            argnum = self.convertargs(fmt, msg["args"], msg["lineno"])
            # build MSGREC
            msgrec = petka.engine.MsgObject(
                len(pe.msgs), msg["wav"], argnum[0], argnum[1], argnum[2])
            msgrec.name = msg["msg"]
            pe.msgs.append(msgrec)

        if destfolder is not None:
            f = open(os.path.join(destfolder, "dialogue.lod"), "wb")
        else:
            f = st_lod
        try:
            pe.write_lod(f)
        finally:
            if destfolder is not None:
                f.close()
        print("DALOGUE.LOD saved: {} messages".\
            format(len(pe.msgs)))

        for grp in compgrp:
            fmt = [("DLGGRP number", self.check32, False),\
                ("DLGGRP argument", self.check32, True)]
            argnum = self.convertargs(fmt, [grp["grp_id"], grp["arg"]], \
                grp["lineno"])
            # build DLGGRP
            grprec = petka.engine.DlgGrpObject(argnum[0], argnum[1])
            grprec.acts = []
            for act in grp["acts"]:
                # act objref
                if act["donref"].upper() == "THIS":
                    donref = grprec.grp_id
                else:
                    donref = self.convertnum(act["donref"])
                if donref is not None:
                    # direct number
                    donref = self.check16(donref, "ON object ref",
                        act["lineno"])
                else:
                    # get by ident
                    self.checkident(act["donref"], "ON object ref",
                        act["lineno"])
                    donref = self.getidentvalue(act["donref"])
                # act arguments
                fmt = []
                for i in range(2):
                    fmt.append(("ON argument {}".format(i + 1),
                        self.check32, True))
                argnum = self.convertargs(fmt, act["args"], act["lineno"])
                # build ON
                actrec = petka.engine.DlgActObject(act["don"], donref,
                    argnum[0], argnum[1])
                actrec.dlgs = []
                for dlg in act["dlgs"]:
                    # dlg arguments
                    fmt = []
                    for i in range(2):
                        fmt.append(("DLG argument {}".format(i + 1),
                            self.check32, True))
                    argnum = self.convertargs(fmt, dlg["args"], dlg["lineno"])
                    # build DLG
                    dlgrec = petka.engine.DlgObject(len(pe.dlgops),
                        argnum[0], argnum[1])
                    for op in dlg["dlgops"]:
                        fmt = [("DLGOP argument", self.check8, False),
                            ("DLGOP ref", self.check16, True)]
                        argnum = self.convertargs(fmt, [op["arg"],
                            op["msg_ref"]], op["lineno"])
                        oprec = petka.engine.DlgOpObject(op["opcode"],
                            argnum[0], argnum[1])
                        pe.dlgops.append(oprec)
                    actrec.dlgs.append(dlgrec)
                grprec.acts.append(actrec)
            pe.dlgs.append(grprec)

        if destfolder is not None:
            f = open(os.path.join(destfolder, "dialogue.fix"), "wb")
        else:
            f = st_fix
        try:
            pe.write_fix(f)
        finally:
            if destfolder is not None:
                f.close()
        print("DALOGUE.FIX saved: {} groups, {} dialog opcodes".\
            format(len(pe.dlgs), len(pe.dlgops)))


    # =======================================================================
    # decompile utils
    # =======================================================================
    def fmtnum16(self, num):
        if num < 10:
            return "{}".format(num)
        elif num == 0xffff:
            return "-1"
        else:
            return "0x{:x}".format(num)

    def fmtnum32(self, num):
        if num < 10:
            return "{}".format(num)
        elif num == 0xffffffff:
            return "-1"
        else:
            return "0x{:x}".format(num)

    def fmtop(self, num):
        if num in petka.OPCODES:
            return petka.OPCODES[num][0]
        return "0x{:x}".format(num)

    def fmtdlgop(self, num):
        if num in petka.DLGOPS:
            return petka.DLGOPS[num][0]
        return "0x{:x}".format(num)

    def escstr(self, value):
        return value.replace("\\", "\\\\").replace("\"", "\\\"")

    # =======================================================================
    # decompile SCRIPT.DAT
    # =======================================================================
    def pretty_print_scr(self, scrname, stream, enc = None, decsort = False):
        def pprint(msg):
            if stream is None:
                print(msg)
            else:
                stream.write((msg + "\n").encode(enc or "UTF-8"))

        pe = petka.Engine()
        pe.init_empty("cp1251")
        bkgname = find_in_folder(os.path.dirname(scrname), "backgrnd.bg")
        resname = find_in_folder(os.path.dirname(scrname), "resource.qrc")
        pe.load_script(scrname, bkgname, resname)

        # define lists of used items
        used_obj = []
        used_res = []

        def fmtfor(num, objid):
            if num == objid:
                return "THIS"
            if num in pe.obj_idx:
                return "obj_{}".format(num)
            if num in pe.scn_idx:
                return "scene_{}".format(num)
            return self.fmtnum16(num)

        def printres(resid):
            pprint("RES res_{} 0x{:x} \"{}\"".format(resid, resid,
                self.escstr(pe.res[resid])))
            pprint("")

        def printrescheck(resid, opcode):
            tp = petka.OPCODES.get(opcode, ("", 0))[1]
            if tp == 1 and resid in pe.res:
                if resid in used_res:
                    return
                used_res.append(resid)
                printres(resid)

        def printitem(item, itemtype):
            pprint("{} {}_{} 0x{:x} \"{}\"".format(itemtype.upper(), itemtype,
                item.idx, item.idx, self.escstr(item.name)))

            # sub objects
            if itemtype == "scene":
                if len(item.refs) == 0:
                    pprint("  ZEROREF")
                for obj, a1, a2, a3, a4, a5 in item.refs:
                    if obj.idx in pe.obj_idx:
                        ref = "obj_{}".format(obj.idx)
                    elif obj.idx in pe.scn_idx:
                        ref = "scene_".format(obj.idx)
                    else:
                        pprint("  # unknown reference to 0x{:x}".format(
                           obj.idx))
                        ref = "0x{:x}".format(obj.idx)
                    pprint("  REF {} {} {} {} {} {}".format(ref,
                        self.fmtnum32(a1), self.fmtnum32(a2),
                        self.fmtnum32(a3), self.fmtnum32(a4),
                        self.fmtnum32(a5)))

            for act in item.acts:
                actif = ""
                if act.act_status != 0xff or act.act_ref != 0xffff:
                    actif = " 0x{:02x} ".format(act.act_status)
                    if act.act_ref == item.idx:
                        actif += "THIS"
                    else:
                        actif += self.fmtnum16(act.act_ref)
                pprint("  ON {}{}".format(self.fmtop(act.act_op), actif))
                # list actions
                for op in act.ops:
                    if op.op_arg1 in pe.res and \
                            petka.OPCODES.get(op.op_code, ["", 0])[1] == 1:
                        res = "res_{}".format(op.op_arg1)
                    else:
                        res = self.fmtnum16(op.op_arg1)
                    pprint("    {} {} {} {} {}".format(self.fmtop(op.op_code),
                        fmtfor(op.op_ref, item.idx), res,
                        self.fmtnum16(op.op_arg2),
                        self.fmtnum16(op.op_arg3)))
                pprint("  ENDON")

            pprint("END{} # {}_{}".format(itemtype.upper(), itemtype, item.idx))
            pprint("")

        pprint("# Decompile SCRIPT \"{}\"".format(scrname))
        pprint("# Version: {}".format(VERSION))
        pprint("# Encoding: {}".format(enc))

        if decsort:
            for idx, scene in enumerate(pe.scenes):
                pprint("# Scene {} / {}".format(idx + 1, len(pe.scenes)))
                # display used objects
                if len(scene.idx.refs) > 0:
                    pprint("# referenced objects {}:".format(len(scene.refs)))
                    for ref in scene.refs:
                        if ref[0].idx in used_obj:
                            pprint("# object 0x{:x} already defined".\
                                format(ref[0].idx))
                            continue
                        used_obj.append(ref[0].idx)
                        if ref[0].idx in pe.obj_idx:
                            for act in ref[0].acts_array:
                                for op in act.ops_array:
                                    printrescheck(op.op_arg1, op.op_code)
                            printitem(obj, "obj")
                else:
                    pprint("# No referenced objects")
                # display res
                for act in scene.acts:
                    for op in act.ops:
                        printrescheck(op.op_arg1, op.op_code)
                printitem(scene, "scene")
            # list unused
            msg = False
            for obj in pe.objects:
                if obj.idx in used_obj: continue
                if not msg:
                    pprint("# Note: Following objects not listed anywhere")
                    msg = True
                printitem(obj, "OBJ")
            msg = False
            for res in self.res:
                if res in used_res: continue
                if not msg:
                    pprint("# Note: Following resources not listed anywhere")
                    msg = True
                printres(res)
        else:
            for obj in pe.objects:
                printitem(obj, "obj")
            for scene in pe.scenes:
                printitem(scene, "scene")
            for res in pe.resord:
                printres(res)

    # =======================================================================
    # decompile DIALOGUE.FIX
    # =======================================================================
    def pretty_print_dlg(self, fixname, stream, enc = None, verbose = False):
        def pprint(msg):
            if stream is None:
                print(msg)
            else:
                stream.write((msg + "\n").encode(enc or "UTF-8"))

        pe = petka.Engine()
        pe.init_empty("cp1251")
        lodname = find_in_folder(os.path.dirname(fixname), "dialogue.lod")
        pe.load_dialogs(fixname, lodname, True)

        pprint("# Decompile DIALOGUE \"{}\"".format(fixname))
        pprint("# Version: {}".format(VERSION))
        pprint("# Encoding: {}".format(enc))

        for msg in pe.msgs:
            pprint("# {} = 0x{:x}".format(msg.idx, msg.idx))
            pprint("MSG msg_{} \"{}\" 0x{:x} 0x{:x} 0x{:x}".format(\
                msg.idx, msg.msg_wav, msg.msg_arg1, msg.msg_arg2, msg.msg_arg3))
            pprint(" \"{}\"".format(self.escstr(msg.name)))
            pprint("")

        for gidx, grp in enumerate(pe.dlgs, 1):
            pprint("# {} = 0x{:x}".format(gidx, gidx))
            pprint("DLGGRP 0x{:x} {}".format(grp.idx, \
                self.fmtnum32(grp.grp_arg1)))
            for sidx, act in enumerate(grp.acts, 1):
                pprint("  ON {} 0x{:x} 0x{:x} 0x{:x} # {}".format(\
                    self.fmtop(act.opcode), act.ref, act.arg1,
                        act.arg2, sidx))
                # print code
                for didx, dlg in enumerate(act.dlgs, 1):
                    #print(bsrec)
                    pprint("    DLG 0x{:x} 0x{:x} # {}".format(\
                        dlg.arg1, dlg.arg2, didx))
                    # scan used addr
                    usedadr = []
                    for op in dlg.ops:
                        if op.opcode == 0x3 or \
                            op.opcode == 0x4: # GOTO or MENURET
                            if op.ref not in usedadr:
                                usedadr.append(op.ref)
                    if len(usedadr) > 0:
                        usedadr.append(dlg.op_start)
                    usedmenu = {}
                    usedcase = {}
                    for oidx, op in enumerate(dlg.ops):
                        cmt = ""
                        opref = "0x{:X}".format(op.ref)
                        opcode = self.fmtdlgop(op.opcode)
                        if op.pos in usedadr:
                             pprint("      label_{:X}:".format(
                                op.pos))
                        if op.opcode == 0x1: # BREAK
                            if op.pos in usedcase:
                                if len(usedadr) > 0:
                                    cmt = "# end select "\
                                        "label_{:X}, case=0x{:}"\
                                        "".format(*usedcase[op.pos])
                                else:
                                    cmt = "# end "\
                                        "select case=0x{:}".\
                                        format(usedcase[op.pos][1])
                        elif op.opcode == 0x2 or op.opcode == 0x8: # MENU or CIRCLE
                            cmt = "# select "
                            doarr = []
                            docurr = []
                            sellen = op.ref % 0x100
                            skiptobrk = False
                            menuactstart = None
                            for oidx2, op2 in enumerate(dlg.ops[oidx + 1:]):
                                if op2.opcode == 0x1: # BREAK
                                    usedcase[op2.pos] = (op.pos, len(doarr))
                                    doarr.append(docurr)
                                    skiptobrk = False
                                    if len(doarr) == sellen:
                                        if op.opcode == 0x2:
                                            menuactstart = oidx2 + oidx + 2
                                        break
                                    docurr = []
                                elif op2.opcode == 0x7 and not skiptobrk: # PLAY
                                    docurr.append("msg_{}".format(op2.ref))
                                else:
                                    docurr = ["complex"]
                            if len(doarr) < sellen:
                                cmt = "# {} select broken, "\
                                    "required={}, got={}".\
                                    format(opcode, sellen, len(doarr))
                            else:
                                cmt += ",".join(["+".join(x) for x in doarr])
                            if menuactstart is not None:
                                for oidx2, op2 in enumerate(dlg.ops[\
                                        menuactstart:menuactstart + sellen]):
                                    usedmenu[op2.pos] = (op.pos, oidx2)
                        elif op.opcode == 0x3 or \
                            op.opcode == 0x4: # GOTO or MENURET
                            opref = "label_{:X}".format(op.ref)
                            if op.pos in usedmenu:
                                cmt = "# action menu="\
                                    "label_{:X}, case=0x{:}".\
                                    format(*usedmenu[op.pos])
                        elif op.opcode == 0x7:
                            opcode = "PLAY"
                            if op.msg:
                                opref = "msg_{}".format(op.ref)
                                if op.ref < len(pe.msgs):
                                    cmt = "# {}".format(self.escstr(
                                        pe.msgs[op.ref].name))
                        oparg = " 0x{:X} ".format(op.arg)
                        if (op.opcode == 0x1 or op.opcode == 0x6) and \
                                op.arg == 0 and op.ref == 0:
                            oparg = ""
                            opref = ""
                        if cmt and verbose:
                            pprint("        " + cmt)
                        pprint("        {}{}{}".\
                            format(opcode, oparg, opref))
                    pprint("    ENDDLG # {}".format(didx))
                pprint("  ENDON # {}".format(sidx))
            pprint("ENDDLGGRP # {}".format(gidx))
            pprint("")


# check if file already exists and flag for overwrite not set
def ckeckoverwrite(fn, args):
    if os.path.exists(fn) and not args.fo:
        print("File \"{}\" already exists, use -fo to overwrite".\
            format(fn))
        return True
    return False

# check files are the same
def checksame(f1, n1, f2, n2):
    if os.path.abspath(f1) == os.path.abspath(f2):
        print("Error: {} and {} are the same \"{}\"".\
            format(f1))
        return True
    return False


def action_dec(args):
    print("Decompile SCRIPT.DAT file")
    destpath = args.destpath
    encoding = args.encoding

    if destpath:
        if ckeckoverwrite(destpath, args): return -1
        if checksame(args.sourcepath, "source", destpath, "destination"):
            return -2
        if not encoding:
            encoding = "UTF-8"

    print("Input:\t{}".format(args.sourcepath))
    print("Output:\t{}".format(destpath or "-"))
    if destpath:
        print("Enc:\t{}".format(encoding))
    if args.decompile_sorted:
        print("Flag decompile_sorted enabled")

    dcs = P12Compiler()
    if destpath:
        f = open(destpath, "wb")
        try:
            dcs.pretty_print_scr(args.sourcepath, f, enc = encoding, \
                decsort = args.decompile_sorted)
        finally:
            f.close()
    else:
        dcs.pretty_print_scr(args.sourcepath, None)

def action_comp(args):
    print("Compile SCRIPT.DAT file")
    print("Input:\t{}".format(args.sourcepath))
    print("Output:\t{}".format(args.destfolder))
    print("Enc:\t{}".format(args.encoding or "UTF-8"))

    dcs = P12Compiler()
    if os.path.exists(args.destfolder) and not args.fo:
        cnt = 0
        lst = ["script.dat", "backgrnd.bg", "resource.qrc"]
        for item in lst:
            if find_in_folder(args.destfolder, item, False):
                print("Error: destination file \"{}\" already "\
                    "exists, use -fo to overwrite".format(item))
                return

    if not os.path.exists(args.destfolder):
        os.makedirs(args.destfolder)
    f = open(args.sourcepath, "rb")
    try:
        dcs.compile_script(f, args.destfolder, args.encoding)
    except ScriptSyntaxError as e:
        if args.trace_error:
            traceback.print_exc()
        print(e, file = sys.stderr)
    finally:
        f.close()

def action_decd(args):
    print("Decompile DIALOGUE.FIX file")
    destpath = args.destpath
    encoding = args.encoding

    if destpath:
        if ckeckoverwrite(destpath, args): return -1
        if checksame(args.sourcepath, "source", destpath, "destination"):
            return -2
        if not encoding:
            encoding = "UTF-8"

    print("Input:\t{}".format(args.sourcepath))
    print("Output:\t{}".format(destpath or "-"))
    if destpath:
        print("Enc:\t{}".format(encoding))
    if args.verbose:
        print("Flag verbose enabled")

    dcs = P12Compiler()
    if destpath:
        f = open(destpath, "wb")
        try:
            dcs.pretty_print_dlg(args.sourcepath, f, enc = encoding, \
                verbose = args.verbose)
        finally:
            f.close()
    else:
        dcs.pretty_print_dlg(args.sourcepath, None, None, \
                verbose = args.verbose)

def action_compd(args):
    print("Compile DIALOGUE.FIX file")
    print("Input:\t{}".format(args.sourcepath))
    print("Output:\t{}".format(args.destfolder))
    print("Enc:\t{}".format(args.encoding or "UTF-8"))

    dcs = P12Compiler()
    if os.path.exists(args.destfolder) and not args.fo:
        cnt = 0
        lst = ["dialogue.fix", "dialogue.lod"]
        for item in lst:
            if find_in_folder(args.destfolder, item, False):
                print("Error: destination file \"{}\" already "\
                    "exists, use -fo to overwrite".format(item))
                return

    if not os.path.exists(args.destfolder):
        os.makedirs(args.destfolder)
    f = open(args.sourcepath, "rb")
    try:
        dcs.compile_dialog(f, args.destfolder, args.encoding)
    except ScriptSyntaxError as e:
        if args.trace_error:
            traceback.print_exc()
        print(e, file = sys.stderr)
    finally:
        f.close()

def internaltest(folder):
    test_arr = [
      "p1demo",
      "p1-0", "p1-1", "p1-2", "p1-3",
      "p2-0", "p2-1", "p2-2"
    ]
    def compare(fn, mem):
        f = open(fn, "rb")
        hf = hashlib.md5()
        try:
            hf.update(f.read())
        finally:
            f.close()
        mem.seek(0)
        hm = hashlib.md5()
        hm.update(mem.read())
        return hm.hexdigest() == hf.hexdigest()

    for test in test_arr:
        print("=== Test: " + test + " ===")
        testbase = os.path.join(folder, test)
        path = find_in_folder(testbase, "script.dat")
        dcs = P12Compiler()
        mems = io.BytesIO()
        dcs.pretty_print_scr(path, mems)
        print("Decompiled script:", mems.tell())
        # compile back
        memscr = io.BytesIO()
        membkg = io.BytesIO()
        memres = io.BytesIO()
        mems.seek(0)
        ndcs = P12Compiler()
        ndcs.compile_script(mems, None, None, \
            memscr, membkg, memres)
        # compare
        if not compare(path, memscr):
            print("SCRIPT.DAT - mismatch")
            break
        if not compare(find_in_folder(testbase, "backgrnd.bg"), membkg):
            print("BACKGRND.BG - mismatch")
            break

        # dialogue
        dcs = P12Compiler()
        path = find_in_folder(testbase, "dialogue.fix")
        if not os.path.exists(path):
            print("All ok - no dialogue")
            continue

        mems = io.BytesIO()
        dcs.pretty_print_dlg(path, mems)
        print("Decompiled dialogue:", mems.tell())
        # compile back
        memfix = io.BytesIO()
        memlod = io.BytesIO()
        mems.seek(0)
        ndcs = P12Compiler()
        ndcs.compile_dialog(mems, None, None, \
            memfix, memlod)
        # compare
        if not compare(path, memfix):
            print("DIALOGUE.FIX - mismatch")
            break
        if not compare(find_in_folder(testbase, "dialogue.lod"), memlod):
            print("DIALOGUE.LOD - mismatch")
            break

        print("All ok")

def action_version(args):
    print("Version: " + VERSION)

def main():
    print(APPNAME + ", " + VERSION)
    print("\tRoman Kharin (romiq.kh@gmail.com)")
    if len(sys.argv) < 2:
        print("Use -h for help.")
        return

    if len(sys.argv) >= 3:
        if sys.argv[1] == "test":
            internaltest(sys.argv[2])
            return

    parser = argparse.ArgumentParser(epilog = \
        "For actions help try: <action> -h")
    subparsers = parser.add_subparsers(title = 'actions')

    # decompile - <script.dat> [[--enc <encoding>] -o <decompiled.txt>]
    parser_dec = subparsers.add_parser("decompile", aliases = ['d'], \
        help = "decompile script.dat")
    parser_dec.add_argument('-fo', action = 'store_true', \
        help = "force overwrite existing output file")
    parser_dec.add_argument('--decompile-sorted', action = 'store_true', \
        help = "display objects and scenes in sorted way (can change order)")
    parser_dec.add_argument('-o', action = 'store', dest = "destpath",\
        help = "output path for decompiled (default: stdout)")
    parser_dec.add_argument('-e', "--enc", action = 'store', dest = "encoding",\
        help = "output encoding (default: UTF-8)")
    parser_dec.add_argument('sourcepath', help = "path to SCRIPT.DAT file")
    parser_dec.set_defaults(func = action_dec)

    # compile - <source.txt> <destination folder> [--enc <encoding>]
    parser_comp = subparsers.add_parser("compile", aliases = ['c'], \
        help = "compile script.dat")
    parser_comp.add_argument('-fo', action = 'store_true', \
        help = "force overwrite existing output files")
    parser_comp.add_argument('-e', "--enc", action = 'store', \
        dest = "encoding", help = "output encoding (default: UTF-8)")
    parser_comp.add_argument('-te', "--trace-error", action = 'store_true', \
        help = "trace syntax error")
    parser_comp.add_argument('sourcepath', help = "path to SOURCE.TXT file")
    parser_comp.add_argument('destfolder', help = "path to output folder")
    parser_comp.set_defaults(func = action_comp)

    # decompiledialog - <dialogue.fix> [[--enc <encoding>] -o <decompiled.txt>]
    parser_decd = subparsers.add_parser("decompiledialog", aliases = ['dd'], \
        help = "decompile dialogue.fix")
    parser_decd.add_argument('-fo', action = 'store_true', \
        help = "force overwrite existing output file")
    parser_decd.add_argument("-v", '--verbose', action = 'store_true', \
        help = "enable more verbose comments")
    parser_decd.add_argument('-o', action = 'store', dest = "destpath",\
        help = "output path for decompiled (default: stdout)")
    parser_decd.add_argument('-e', "--enc", action = 'store', \
        dest = "encoding", help = "output encoding (default: UTF-8)")
    parser_decd.add_argument('sourcepath', help = "path to DIALOGUE.FIX file")
    parser_decd.set_defaults(func = action_decd)

    # compiledialog - <source.txt> <destination folder> [--enc <encoding>]
    parser_compd = subparsers.add_parser("compiledialog", aliases = ['cd'], \
        help = "compile dialogue.fix")
    parser_compd.add_argument('-fo', action = 'store_true', \
        help = "force overwrite existing output files")
    parser_compd.add_argument('-e', "--enc", action = 'store', \
        dest = "encoding", help = "output encoding (default: UTF-8)")
    parser_compd.add_argument('-te', "--trace-error", action = 'store_true', \
        help = "trace syntax error")
    parser_compd.add_argument('sourcepath', help = "path to SOURCE.TXT file")
    parser_compd.add_argument('destfolder', help = "path to output folder")
    parser_compd.set_defaults(func = action_compd)

    # version
    parser_version = subparsers.add_parser("version", help = "program version")
    parser_version.set_defaults(func = action_version)

    args = parser.parse_args()
    args.func(args)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# romiq.kh@gmail.com, 2014

import os
import struct
import io

from .fman import FileManager
from . import EngineError

OPCODES = {
    1:  ("USE",         0),
    2:  ("SETPOS",      2),
    3:  ("GOTO",        0),
    4:  ("LOOK",        0),
    5:  ("SAY",         0),
    6:  ("TAKE",        0),
    9:  ("WALK",        2),
    10: ("TALK",        0),
    11: ("END",         0),
    14: ("SET",         1),
    15: ("SHOW",        1),
    16: ("HIDE",        0),
    17: ("DIALOG",      1),
    18: ("ZBUFFER",     0),
    19: ("TOTALINIT",   1),
    20: ("ANIMATE",     1),
    21: ("STATUS",      1),
    22: ("ADDINV",      0),
    23: ("DELINV",      0),
    24: ("STOP",        1),
    25: ("CURSOR",      1),
    26: ("OBJECTUSE",   0),
    27: ("ACTIVE",      1),
    28: ("SAID",        0),
    29: ("SETSEQ",      0),
    30: ("ENDSEQ",      0),
    31: ("CHECK",       0),
    32: ("IF",          0),
    33: ("DESCRIPTION", 0),
    34: ("HALF",        0),
    36: ("WALKTO",      0),
    37: ("WALKVICH",    0),
    38: ("INITBG",      0),
    39: ("USERMSG",     0),
    40: ("SYSTEM",      0),
    41: ("SETZBUFFER",  0),
    42: ("CONTINUE",    0),
    43: ("MAP",         1),
    44: ("PASSIVE",     1),
    45: ("NOMAP",       1),
    46: ("SETINV",      1),
    47: ("BGSFX",       1),
    48: ("MUSIC",       1),
    49: ("IMAGE",       1),
    50: ("STAND",       1),
    51: ("ON",          1),
    52: ("OFF",         1),
    53: ("PLAY",        1),
    54: ("LEAVEBG",     0),
    55: ("SHAKE",       1),
    56: ("SP",          2),
    57: ("RANDOM",      1),
    58: ("JUMP",        0),
    59: ("JUMPVICH",    0),
    60: ("PART",        2),
    61: ("CHAPTER",     2),
    62: ("AVI",         1),
    63: ("TOMAP",       0),
}

DLGOPS = {
    1:  ("BREAK",       0),
    2:  ("MENU",        1),
    3:  ("GOTO",        3),
    4:  ("MENURET",     4),
    6:  ("RETURN",      0),
    7:  ("PLAY",        5),
    8:  ("CIRCLE",      6),
}

ACTIONS = {
    0: ("VIEW",    5002),
    1: ("GOTO",    5003),
    2: ("USE",     5004),
    3: ("TAKE",    5005),
    4: ("TALK",    5006),
    5: ("CHAPAY",  5007),
    6: ("INVNTR",    -1),

}

class ScrObject:
    def __init__(self, idx, name):
        self.idx = idx
        self.name = name
        self.acts = None # action hadlers
        self.cast = None # object color (CASTS.INI)

class ScrActObject:
    def __init__(self, act_op, act_status, act_ref):
        self.act_op = act_op         # handler: opcode filter
        self.act_status = act_status # handler: status filter
        self.act_ref = act_ref       # handler: object idx filter
        self.ops = None              # operations

class ScrOpObject:
    def __init__(self, op_ref, op_code, op_arg1, op_arg2, op_arg3):
        self.op_ref = op_ref # object idx
        self.op_code = op_code # opcode
        self.op_arg1 = op_arg1 # resource id, etc
        self.op_arg2 = op_arg2
        self.op_arg3 = op_arg3

class ScrObjectState:
    def __init__(self, obj):
        self.obj = obj
        self.state = 0
        self.prop = [0] * 8

class MsgObject:
    def __init__(self, idx, wav, arg1, arg2, arg3):
        self.idx = idx
        self.msg_wav = wav   # wav filename
        self.msg_arg1 = arg1 # reference to object
        self.msg_arg2 = arg2
        self.msg_arg3 = arg3
        self.name = None

class DlgGrpObject:
    def __init__(self, idx, arg1):
        self.idx = idx
        self.grp_arg1 = arg1
        self.acts = None # dialog handlers

class DlgActObject:
    def __init__(self, opcode, ref, arg1, arg2):
        self.opcode = opcode # handler: opcode filter
        self.ref = ref       # handler: object idx filter
        self.arg1 = arg1
        self.arg2 = arg2
        self.dlgs = None     # dialogs
        self.obj = None      # handler object

class DlgObject:
    def __init__(self, op_start, arg1, arg2):
        self.op_start = op_start # start position
        self.arg1 = arg1
        self.arg2 = arg2
        self.ops = None          # operations list

class DlgOpObject:
    def __init__(self, opcode, arg, ref):
        self.opcode = opcode    # dialog opcode
        self.arg = arg          # argument (ref, offset etc.)
        self.ref = ref          # message idx
        self.msg = None         # message
        self.pos = None

class Engine:
    def __init__(self):
        self.fman = None
        self.parts = []
        self.start_part = None
        self.start_chap = None
        self.start_scene = None

        self.curr_part = None
        self.curr_chap = None

        self.curr_path = None
        self.curr_speech = None
        self.curr_diskid = None

    def init_empty(self, enc):
        self.enc = enc
        self.objects = []
        self.scenes = []
        self.obj_idx = {} # id -> object
        self.scn_idx = {} # id -> scene
        self.msgs = []
        self.dlgs = []
        self.dlg_idx = {}
        self.dlgops = []

    def parse_ini(self, f):
        # parse ini settings
        curr_sect = None
        ini = {}
        order_sect = []
        orders = {}
        for line in f.readlines():
            line = line.decode(self.enc).strip()
            if len(line) == 0: continue
            if line[:1] == ";": continue
            if line[:1] == "[" and line[-1:] == "]":
                if curr_sect is not None:
                    orders[curr_sect] = order
                curr_sect = line[1:-1].strip()
                order_sect.append(curr_sect)
                order = []
                ini[curr_sect] = {}
                continue
            kv = line.split("=", 1)
            if len(kv) != 2: continue
            ini[curr_sect][kv[0].strip()] = kv[1].strip()
            order.append(kv[0].strip())
        orders[curr_sect] = order
        ini["__ordersect__"] = order_sect
        ini["__order__"] = orders
        return ini

    def parse_res(self, f):
        res  = {}
        resord = []
        for line in f.readlines():
            line = line.decode(self.enc).strip()
            if len(line) == 0:
                continue
            pair = line.split("=", 1)
            if len(pair) < 2:
                continue
            value = pair[1].strip()
            if value[:1] == "=":
                value = value[1:].strip()
            res_id = int(pair[0].strip(), 10)
            res[res_id] = value
            resord.append(res_id)
        return res, resord

    def load_data(self, folder, enc):
        self.init_empty(enc)
        self.fman = FileManager(folder)
        # load PARTS.INI
        pf = self.fman.find_path("parts.ini")
        if pf:
            f = open(pf, "rb")
            try:
                self.parts_ini = self.parse_ini(f)
            finally:
                f.close()
            for sect in self.parts_ini["__ordersect__"]:
                data = self.parts_ini[sect]
                if sect == "All":
                    if "Part" in data:
                        self.start_part = int(data["Part"])
                    if "Chapter" in data:
                        self.start_chap = int(data["Chapter"])
                elif sect[:5] == "Part ":
                    self.parts.append(sect)
        else:
            # load BGS.INI only (e.g. DEMO)
            self.parts_ini = None

        # std stores
        self.fman.load_store("patch.str")
        self.fman.load_store("main.str")

    def open_part(self, part, chap):
        self.fman.unload_stores(1)
        self.curr_part = part
        self.curr_chap = chap
        if self.parts_ini:
            pname = "Part {}".format(part)
            pcname = pname
            if chap:
                pcname += " Chapter {}".format(chap)
            ini = self.parts_ini[pname]
            self.curr_path = ini["CurrentPath"]
            self.curr_speech = ini["PathSpeech"]
            self.curr_diskid = ini["DiskID"]
            inic = self.parts_ini[pcname]
            if "Chapter" in inic:
                self.fman.load_store(inic["Chapter"], 1)
        else:
            ini = {}
            self.curr_path = ""
            self.curr_speech = ""
            self.curr_diskid = None

        # load .STR
        strs = ["Flics", "Background", "Wav", "Music", "SFX", "Speech"]
        for strf in strs:
            if strf in ini:
                self.fman.load_store(ini[strf], 1)
        # load script.dat, backgrnd.bg, resources.qrc, etc
        self.load_script()
        # load persp & scenes enter points
        self.load_bgs()
        # load names & invntr
        self.load_names()
        # load dialogs
        self.load_dialogs()

        # current state
        self.curr_scene = None
        self.curr_obj = None
        self.curr_dlg = None
        self.curr_cursor = None
        self.curr_char1 = None
        self.curr_char2 = None
        self.curr_invntr = None

    def load_script(self, scrname = None, bkgname = None, resname = None):
        self.objects = []
        self.scenes = []
        self.obj_idx = {}
        self.scn_idx = {}

        if scrname is None:
            try:
                data = self.fman.read_file(self.curr_path + "script.dat")
            except:
                raise EngineError("Can't open SCRIPT.DAT")
        else:
            try:
                f = open(scrname, "rb")
            except:
                raise EngineError("Can't open SCRIPT.DAT")
            try:
                data = f.read()
            finally:
                f.close()
        num_obj, num_scn = struct.unpack_from("<II", data[:8])
        off = 8
        def read_rec(off):
            obj_id, name_len = struct.unpack_from("<HI", data[off:off + 6])
            off += 6
            name = data[off:off + name_len].decode(self.enc)
            off += name_len
            num_act = struct.unpack_from("<I", data[off:off + 4])[0]
            off += 4
            acts = []
            for i in range(num_act):
                act_op, act_status, act_ref, num_op = struct.unpack_from(\
                    "<HBHI", data[off:off + 9])
                off += 9
                act = ScrActObject(act_op, act_status, act_ref)
                act.ops = []
                for j in range(num_op):
                    op = struct.unpack_from("<5H", data[off:off + 10])
                    off += 10
                    op = ScrOpObject(*op)
                    act.ops.append(op)
                acts.append(act)
            rec = ScrObject(obj_id, name)
            rec.acts = acts
            return off, rec

        for i in range(num_obj):
            off, obj = read_rec(off)
            self.objects.append(obj)
            self.obj_idx[obj.idx] = obj

        for i in range(num_scn):
            off, scn = read_rec(off)
            scn.refs = None
            self.scenes.append(scn)
            self.scn_idx[scn.idx] = scn

        if bkgname is None:
            try:
                data = self.fman.read_file(self.curr_path + "backgrnd.bg")
            except:
                data = None
        else:
            try:
                f = open(bkgname, "rb")
            except:
                data = None
            try:
                data = f.read()
            finally:
                f.close()

        if data:
            num_rec = struct.unpack_from("<I", data[:4])[0]
        else:
            num_rec = 0
        off = 4
        for i in range(num_rec):
            scn_ref, num_ref = struct.unpack_from("<HI", data[off:off + 6])
            off += 6
            if scn_ref in self.scn_idx:
                scn = self.scn_idx[scn_ref]
                scn.refs = []
            else:
                raise EngineError("DEBUG: Scene ID = 0x{:x} not found".\
                    format(scn_ref))

            for j in range(num_ref):
                ref = struct.unpack_from("<H5I", data[off:off + 22])
                off += 22
                if ref[0] in self.obj_idx:
                    obj = self.obj_idx[ref[0]]
                    scn.refs.append([obj] + list(ref[1:]))
                else:
                    raise EngineError("DEBUG: Scene ref 0x{:x} not found".\
                        format(obj[0]))

        if resname is None:
            try:
                f = self.fman.read_file_stream(self.curr_path + "resource.qrc")
            except:
                f = None
        else:
            try:
                f = open(resname, "rb")
            except:
                f = None
        try:
            if f:
                self.res, self.resord = self.parse_res(f)
            else:
                self.res = {}
                self.resord = []
        finally:
            if f:
                f.close()

    def load_names(self):
        self.names = {}
        self.namesord = []
        fp = self.curr_path + "names.ini"
        if self.fman.exists(fp):
            f = self.fman.read_file_stream(fp)
            ini = self.parse_ini(f)
            self.names = ini["all"]
            self.namesord = ini["__order__"]["all"]
            f.close()

        self.invntr = {}
        self.invntrord = []
        fp = self.curr_path + "invntr.txt"
        if self.fman.exists(fp):
            f = self.fman.read_file_stream(fp)
            ini = self.parse_ini(f)
            self.invntr = ini["ALL"]
            self.invntrord = ini["__order__"]["ALL"]
            f.close()

        self.casts = {}
        self.castsord = []
        fp = self.curr_path + "cast.ini"
        if self.fman.exists(fp):
            f = self.fman.read_file_stream(fp)
            ini = self.parse_ini(f)
            self.casts = ini["all"]
            self.castsord = ini["__order__"]["all"]
            f.close()

        # bind casts to objects
        for obj in self.objects:
            if obj.name in self.casts:
                # parse color
                try:
                    val = self.casts[obj.name].split(" ")
                    val = [x for x in val if x]
                    r = int(val[0])
                    g = int(val[1])
                    b = int(val[2])
                except:
                    r, g, b = 255, 255, 255
                obj.cast = (r, g, b)

    def load_bgs(self):
        # load BGS.INI
        self.bgs_ini = {}
        self.start_scene = None
        bgsfn = self.curr_path + "bgs.ini"
        if self.fman.exists(bgsfn):
            f = self.fman.read_file_stream(bgsfn)
            try:
                self.bgs_ini = self.parse_ini(f)
            finally:
                f.close()

        settings = self.bgs_ini.get("Settings", {})
        self.start_scene = settings.get("StartRoom", None)

        # bgs.ini: parse enter areas and perspective
        for scene in self.scenes:
            scene.entareas = None
            areas = self.bgs_ini.get(scene.name, {})
            if scene.name in self.bgs_ini:
                scene.entareas = []
                for key in areas.keys():
                    # search scene
                    sf = None
                    for scenefrom in self.scenes:
                        if scenefrom.name == key:
                            sf = scenefrom
                            break
                    value = areas[key]
                    # search objects
                    oo = None
                    for objon in self.objects:
                        if objon.name == value:
                            oo = objon
                            break
                    if sf and oo:
                        scene.entareas.append((sf, oo))
            # persp
            persp = settings.get(scene.name, "")
            persp = persp.split(" ")
            persp = [x for x in persp if x]
            try:
                persp = [float(persp[0]), float(persp[1]),
                    int(persp[2]), int(persp[3]), float(persp[4])]
            except:
                persp = None
            scene.persp = persp


    def load_dialogs(self, fixname = None, lodname = None, noobjref = False):
        self.msgs = []
        # DIALOGUES.LOD

        if lodname is None:
            try:
                f = self.fman.read_file_stream(self.curr_path + "dialogue.lod")
            except:
                f = None
        else:
            try:
                f = open(lodname, "rb")
            except:
                f = None

        if f:
            try:
                temp = f.read(4)
                num_msg = struct.unpack_from("<I", temp)[0]
                for i in range(num_msg):
                    temp = f.read(24)
                    arg1, wav, arg2, arg3 = struct.unpack_from("<I12sII", temp)
                    msg = MsgObject(len(self.msgs), \
                        wav.decode(self.enc).strip(), arg1, arg2, arg3)
                    # scan objects
                    if not noobjref:
                        msg.obj = self.obj_idx.get(arg1, None)
                        if not msg.obj:
                            raise EngineError("DEBUG: Message ref = 0x{:x} not found".\
                            format(arg1))
                    self.msgs.append(msg)
                for i, capt in enumerate(f.read().split(b"\x00")):
                    if i < len(self.msgs):
                        self.msgs[i].name = capt.decode(self.enc)
            finally:
                if f:
                    f.close()

        # DIALOGUES.FIX
        self.dlgs = []
        self.dlg_idx = {}
        self.dlgops = []
        if fixname is None:
            try:
                f = self.fman.read_file_stream(self.curr_path + "dialogue.fix")
            except:
                f = None
        else:
            try:
                f = open(fixname, "rb")
            except:
                f = None

        if f:
            try:
                temp = f.read(4)
                num_grps = struct.unpack_from("<I", temp)[0]
                for i in range(num_grps):
                    temp = f.read(12)
                    idx, num_acts, arg1 = struct.unpack_from("<III", temp)
                    grp = DlgGrpObject(idx, arg1)
                    grp.num_acts = num_acts
                    self.dlgs.append(grp)
                opref = {}
                for grp in self.dlgs:
                    self.dlg_idx[grp.idx] = grp
                    grp.acts = []
                    for i in range(grp.num_acts):
                        temp = f.read(16)
                        opcode, ref, num_dlgs, arg1, arg2 = \
                            struct.unpack_from("<2H3I", temp)
                        act = DlgActObject(opcode, ref, arg1, arg2)
                        act.num_dlgs = num_dlgs
                        if not noobjref:
                            if ref not in self.obj_idx:
                                raise EngineError("Dialog group 0x{:x} refered "\
                                    "to unexisted object 0x{:x}".format(
                                    grp.idx, ref))
                            act.obj = self.obj_idx[act.ref]
                        grp.acts.append(act)
                    for act in grp.acts:
                        act.dlgs = []
                        for i in range(act.num_dlgs):
                            temp = f.read(12)
                            op_start, arg1, arg2 = \
                                struct.unpack_from("<3I", temp)
                            if op_start in opref:
                                raise EngineError(
                                    "Multiple dialog opcodes reference")
                            dlg = DlgObject(op_start, arg1, arg2)
                            opref[op_start] = dlg
                            dlg.ops = None
                            act.dlgs.append(dlg)
                temp = f.read(4)
                num_ops = struct.unpack_from("<I", temp)[0]
                for oidx, i in enumerate(range(num_ops)):
                    temp = f.read(4)
                    ref, arg, code = struct.unpack_from("<HBB", temp)
                    dlgop = DlgOpObject(code, arg, ref)
                    dlgop.pos = oidx
                    if ref < len(self.msgs):
                        dlgop.msg = self.msgs[ref]
                    self.dlgops.append(dlgop)
                dlg = None
                oparr = []
                for idx, oprec in enumerate(self.dlgops):
                    if idx in opref:
                        if len(oparr) > 0:
                            dlg.ops = oparr
                            oparr = []
                        dlg = opref[idx]
                    oparr.append(oprec)
                if len(oparr) > 0:
                    dlg.ops = oparr
            finally:
                if f:
                    f.close()

    def write_script(self, f):
        f.write(struct.pack("<II", len(self.objects), len(self.scenes)))

        def write_rec(rec):
            ename = rec.name.encode(self.enc)
            f.write(struct.pack("<HI", rec.idx, len(ename)))
            f.write(ename)
            f.write(struct.pack("<I", len(rec.acts)))
            for act in rec.acts:
                f.write(struct.pack("<HBHI", act.act_op, act.act_status,
                    act.act_ref, len(act.ops)))
                for op in act.ops:
                    f.write(struct.pack("<5H", op.op_ref, op.op_code,
                        op.op_arg1, op.op_arg2, op.op_arg3))

        for rec in self.objects + self.scenes:
            write_rec(rec)

    def write_backgrnd(self, f):
        lst = [scn for scn in self.scenes if scn.refs is not None]
        f.write(struct.pack("<I", len(lst)))
        for scn in lst:
            f.write(struct.pack("<HI", scn.idx, len(scn.refs)))
            for ref in scn.refs:
                f.write(struct.pack("<H5I", ref[0].idx, ref[1], ref[2],
                    ref[3], ref[4], ref[5]))


    def write_lod(self, f):
        f.write(struct.pack("<I", len(self.msgs)))
        for msg in self.msgs:
            wav = msg.msg_wav.encode(self.enc)
            while len(wav) < 12:
                wav += b"\0"
            f.write(struct.pack("<I12sII", msg.msg_arg1, wav, msg.msg_arg2,
                msg.msg_arg3))
        for msg in self.msgs:
            txt = msg.name.encode(self.enc)
            f.write(txt + b"\0")

    def write_fix(self, f):
        f.write(struct.pack("<I", len(self.dlgs)))
        for grp in self.dlgs:
            f.write(struct.pack("<3I", grp.idx, len(grp.acts), grp.grp_arg1))
        for grp in self.dlgs:
            for act in grp.acts:
                f.write(struct.pack("<2H3I", act.opcode, act.ref,
                    len(act.dlgs), act.arg1, act.arg2))
            for act in grp.acts:
                for dlg in act.dlgs:
                    f.write(struct.pack("<3I", dlg.op_start, dlg.arg1,
                        dlg.arg2))
        f.write(struct.pack("<I", len(self.dlgops)))
        for op in self.dlgops:
            f.write(struct.pack("<H2B", op.ref, op.arg, op.opcode))

    def scene_to_id(self, name):
        pass

    # create current state from initial state
    def init_game(self):
        self.curr_scene = self.scene_to_id(self.start_scene)

        self.curr_obj = []
        for obj in self.objects + self.scenes:
            state = ScrObjectState(obj)
            self.curr_obj.append(state)
            state.prop

        self.curr_dlg = []
        for dlgop in self.dlgops:
            dlgstate = DlgOpObject(dlgop.code, dlgop.arg, dlgop.ref)
            dlgstate.pos = dlgop.pos
            self.curr_dlg.append(dlgstate)

        self.curr_cursor = None
        self.curr_char1 = None
        self.curr_char2 = None
        self.curr_invntr = []

    def load_save(self, ls):
        self.curr_scene = self.scene_to_id(ls.scene)

        self.curr_obj = None
        self.curr_dlg = None
        self.curr_cursor = None
        self.curr_char1 = None
        self.curr_char2 = None
        self.curr_invntr = None



from cx_Freeze import setup, Executable

import cx_Freeze.util

import sys, os, struct

# Dependencies are automatically detected, but it might need
# fine tuning.
buildOptions = dict(packages = ["re", "io", "PIL", "traceback", "zlib", "gzip",
        "argparse", "struct", "binascii"], \
    excludes = ["_posixsubprocess"],
    include_files = ["help"],
    compressed = True, silent = True,\
    optimize = 2, copy_dependent_files = True, \
    create_shared_zip = True, include_in_shared_zip = True)

executables = [
    Executable('p12explore.py',
        base = 'Win32GUI',
        targetName = "p12explore.exe"),
    Executable('p12script.py',
        base = 'Console',
        targetName = "p12script.exe")
]

setup(name='p12explore',
      version = '0.3',
      description = 'Petka 1&2 utilities',
      author = "romiq.kh@gmail.com, https://bitbucket.org/romiq/p12simtran",
      options = dict(build_exe = buildOptions),
      executables = executables)




class EngineError(Exception): pass

from .engine import Engine, OPCODES, DLGOPS, ACTIONS
from .fman import FileManager
from .imgbmp import BMPLoader
from .imgflc import FLCLoader
from .imgleg import LEGLoader
from .imgmsk import MSKLoader
from .saves import SaveLoader



# -*- coding: utf-8 -*-

# romiq.kh@gmail.com, 2014

import array, struct, io

from . import EngineError

try:
    from PIL import Image
except ImportError:
    Image = None

try:
    from PIL import FliImagePlugin
except ImportError:
    pass

FLC_HEADER = [
    ["fsize",        1, "I", False],
    ["ftype",        1, "H", False],
    ["frames_num",   1, "H", True],
    ["width",        1, "H", True],
    ["height",       1, "H", True],
    ["depth",        1, "H", False],
    ["flags",        1, "H", True],
    ["speed",        1, "I", True],
    ["reserved1",    1, "H", False],
    ["created",      1, "I", True],
    ["creator",      1, "I", True],
    ["updated",      1, "I", True],
    ["updater",      1, "I", True],
    ["aspect_dx",    1, "H", True],
    ["aspect_dy",    1, "H", True],
    ["ext_flags",    1, "H", False],
    ["keyframes",    1, "H", False],
    ["totalframes",  1, "H", False],
    ["req_memory",   1, "I", False],
    ["max_regions",  1, "H", False],
    ["transp_num",   1, "H", False],
    ["reserved2",   24, "s", False],
    ["oframe1",      1, "I", False],
    ["oframe2",      1, "i", False],
    ["reserved3",   40, "s", False],
]

class FLCLoader:
    def __init__(self):
        self.rgb = None
        self.image = None
        self.width = 0
        self.height = 0
        self.frame_num = 0
        self.delay = 0


    def load_info(self, f):
        self.image = Image.open(f)
        self.frame_num = 1
        try:
            while 1:
                self.image.seek(self.image.tell() + 1)
                self.frame_num += 1
        except EOFError:
            pass # end of sequence


    def parseflcchunks(self, f, offset, limit, level = 0, maxchunks = None,):
        def check_hdr(size, delta, name, offset):
            if delta < size:
                raise EngineError("Incorrect FLC %s chunk at 0x{:08x}".format(
                    (name, offset)))

        chunks = []
        while True:
            if limit is not None:
                if offset >= limit:
                    break
            if maxchunks is not None:
                if len(chunks) >= maxchunks:
                    break
            chunk = {"offset": offset}
            temp = f.read(6)
            sz, tp = struct.unpack_from("<IH", temp)
            offset += 6

            chunk["size"] = sz
            chunk["type"] = tp
            delta = sz - 6
            if delta < 0:
                raise EngineError("Incorrect FLC chunk at 0x{:08x}".format(
                    chunk["offset"]))

            raw_chunks = [
                0x4, # COLOR_256
                0x7, # DELTA_FLC
                0xf, # BYTE_RUN
                0xF100, # PREFIX_TYPE - mismaked, out destination ignore this
            ]
            #print("{}CHUNK 0x{:x}, size = {}".format("  "*level, tp, sz))
            # do not parse 3rd level 0x12 chunk
            if tp == 0x12 and level == 2:
                tp = 0x4

            # parse chunks
            if tp in raw_chunks:
                temp = f.read(delta)
                offset += delta
                chunk["data"] = temp
            elif tp == 0x12:
                # PSTAMP
                check_hdr(6, delta, "PSTAMP", offset)
                temp = f.read(6)
                delta -= 6
                height, width, xlate = struct.unpack_from("<3H", temp)
                offset += 6
                offset, subchunks = self.parseflcchunks(f, offset,
                    offset + delta, level + 1, 1)
                chunk["chunks"] = subchunks
                #print(subchunks)
            elif tp == 0xF1FA:
                # FRAME_TYPE
                check_hdr(10, delta, "FRAME_TYPE", offset)
                temp = f.read(10)
                delta -= 10
                sub_num, delay, reserved, width, height = \
                    struct.unpack_from("<5H", temp)
                offset += 10
                chunk["delay"] = delay
                chunk["width"] = width
                chunk["height"] = height
                offset, subchunks = self.parseflcchunks(f, offset,
                    offset + delta, level + 1, sub_num)
                chunk["chunks"] = subchunks
            else:
                raise Exception("Unknown FLC chunk type 0x{:04x} at 0x{:x08x}".\
                    format(tp, offset))

            chunks.append(chunk)

        return offset, chunks

    def load_data(self, f):
        # parse header
        offset = 0
        hdr_keys = []
        hdr_struct = "<"
        for hnam, hsz, htp, hed in FLC_HEADER:
            hdr_keys.append(hnam)
            if hsz == 1:
                hdr_struct += htp
            else:
                hdr_struct += "%d" % hsz + htp

        header = {}
        temp = f.read(128)
        hdr = struct.unpack_from(hdr_struct, temp)

        offset += 128

        if len(hdr) != len(hdr_keys):
            raise EngineError("Incorrect FLC header {} != {}".format(
                len(hdr), len(hdr_keys)))
        for hid in range(len(hdr)):
            header[hdr_keys[hid]] = hdr[hid]

        if header["ftype"] != 0xAF12:
            raise EngineError("Unsupported FLC type (0x{:04x})".format(
                header["ftype"]))

        # check if not EGI ext
        if header["creator"] == 0x45474900:
            if header["ext_flags"] != 0:
                raise EngineError("Unsupported FLC EGI extension")

        # NOTE: we recreate FLC to avoid Pilllow bug
        #  1. remove 0xf100 chunk  (PREFIX, implementation specific)
        #  2. remobe 0x12 subchunk (PSTAMP) from 1st frame

        # read chunks
        _, chunks = self.parseflcchunks(f, offset, header["fsize"])

        f.seek(0)
        buf = io.BytesIO()
        buf.write(f.read(128)) # clone header
        for chunk in chunks:
            if chunk["type"] == 0xF100:
                continue
            elif chunk["type"] == 0xF1FA:
                rebuild = False
                nchunks = []
                nsz = 16 # I6H - type, size, sub_num, delay,
                         # reserved, width, height
                for schunk in chunk["chunks"]:
                    if schunk["type"] == 0x12: # detect mailformed PSTAMP
                        rebuild = True
                    elif rebuild:
                        nchunks.append(schunk)
                        nsz += schunk["size"]
                if rebuild:
                    buf.write(struct.pack("<I6H", nsz, 0xF1FA, len(nchunks),
                        chunk["delay"], 0, chunk["width"], chunk["height"]))
                    for schunk in nchunks:
                        f.seek(schunk["offset"])
                        buf.write(f.read(schunk["size"]))
                    continue
            # copy chunk
            buf.write(f.read(chunk["size"]))

        buf.seek(0)
        self.image = Image.open(buf)



# -*- coding: utf-8 -*-

# romiq.kh@gmail.com, 2014

import array, struct, io

from . import EngineError

try:
    from PIL import Image
except ImportError:
    Image = None

class BMPLoader:
    def __init__(self):
        self.rgb = None
        self.image = None
        self.width = 0
        self.height = 0

    def load_data_int16(self, f):
        # check magic string "BM"
        temp = f.read(2)
        if temp != b"BM":
            raise EngineError("Bad magic string")
        off = 2

        temp = f.read(12)
        f_sz, res1, res2, data_offset = struct.unpack_from("<IHHI", temp)
        off += 12

        # read next 40 bytes, BITMAPINFOHEADER
        temp = f.read(40)
        pict = struct.unpack_from("<IiiHHIIiiII", temp)
        off += 40
        if pict[0] != 40:
            raise EngineError("Unsupported InfoHeader")
        pictw = pict[1]
        picth = pict[2]

        # read data_offset - 40 - 6 bytes
        delta = data_offset - 40 - 6
        if delta < 0:
            raise EngineError("To small bitmap data offset")
        if delta != 8:
            raise EngineError("Unsupported Header at 0x36")
        temp = f.read(delta)
        hdr36 = struct.unpack_from("<II", temp)
        off += delta

        bsz = pictw * picth * 2
        picture_data = f.read(bsz)

        off += bsz
        if len(picture_data) != bsz:
            raise EngineError("Bitmap truncated, need {}, got {}".format(bsz, \
                len(picture_data)))

        # read 2 zero bytes
        #temp = f.read(2)
        #if temp != b"\x00\x00":
        #    raise EngineError("Magic zero bytes absent or mismatch")
        #off += 2

        temp = f.read(2)
        if len(temp) > 0:
            if temp != b"\x00\x00":
                raise EngineError("BMP read error, some data unparsed")

        return pictw, picth, picture_data

    def pixelswap16ud(self, pw, ph, pd):
        # convert 16 bit to 24 + vertical reverse
        b16arr = array.array("H") # unsigned short
        b16arr.frombytes(pd)
        b16arr.byteswap()
        rgb = array.array("B", [0] * pw * ph * 3)
        for j in range(ph):
            for i in range(pw):
                off = (ph - j - 1) * pw * 3 + i * 3
                b16 = b16arr[j * pw + i]
                rgb[off] = (b16 << 3) & 0b11111000
                rgb[off + 1] = (b16 >> 3) & 0b11111100
                rgb[off + 2] = (b16 >> 8) & 0b11111000
        return rgb

    def pixelswap16(self, pw, ph, pd):
        # convert 16 bit to 24
        b16arr = array.array("H") # unsigned short
        b16arr.frombytes(pd)
        #b16arr.byteswap()
        rgb = array.array("B", [0] * pw * ph * 3)
        for j in range(ph):
            for i in range(pw):
                off = j * pw * 3 + i * 3
                b16 = b16arr[j * pw + i]
                rgb[off + 2] = (b16 << 3) & 0b11111000
                rgb[off + 1] = (b16 >> 3) & 0b11111100
                rgb[off + 0] = (b16 >> 8) & 0b11111000
        return rgb

    def load_info(self, f):
        try:
            pw, ph, pd = self.load_data_int16(f)
            self.width = pw
            self.height = ph
        except:
            f.seek(0)
            self.image = Image.open(f)

    def load_raw(self, pw, ph, pd):
        if Image:
            pd = self.pixelswap16(pw, ph, pd).tobytes()
            self.image = Image.frombytes("RGB", (pw, ph), pd)
        else:
            self.width = pw
            self.height = ph
            self.rgb = self.pixelswap16(pw, ph, pd)

    def load_data(self, f):
        try:
            pw, ph, pd = self.load_data_int16(f)
            if Image:
                pd = self.pixelswap16ud(pw, ph, pd).tobytes()
                self.image = Image.frombytes("RGB", (pw, ph), pd)
            else:
                self.width = pw
                self.height = ph
                self.rgb = self.pixelswap16(pw, ph, pd)
        except:
            f.seek(0)
            self.image = Image.open(f)
